package org.processmining.humbertogarciacaballero.models.gui.components;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingConstants;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.gui.GraphVisualization;
import org.processmining.humbertogarciacaballero.models.gui.images.ImageHelper;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.Constants;
import prefuse.Display;
import prefuse.Visualization;
import prefuse.data.Node;
import prefuse.data.Tuple;
import prefuse.data.expression.parser.ExpressionParser;
import prefuse.data.tuple.TupleSet;
import prefuse.data.util.BreadthFirstIterator;
import prefuse.visual.VisualItem;

public class SoundnessInteractionGUI extends JPanel {

	private PNetVisualizerController controller;

	private Visualization vis;

	private int current;
	private int maxDepth;

	private TupleSet focusGroupHighlight;
	private TupleSet focusGroupFit;

	private enum Mode {
		NON_LIVE, DEAD, UNBOUNDED;
	}

	private Mode mode;

	private boolean changesMade;
	
	private BreadthFirstIterator bfi;

	// GUI
	private JButton rightButton;
	private JButton leftButton;
	private JLabel label;

	/**
	 * Create the frame.
	 */
	public SoundnessInteractionGUI(Visualization vis, Display display) {
		this.vis = vis;
		current = -1;
		maxDepth = -1;
		controller = PNetVisualizerController.getInstance();
		focusGroupHighlight = vis.getFocusGroup(GraphVisualization.GRAPH_NODES_PROBLEMATIC_HIGHLIGHT);
		focusGroupFit = vis.getFocusGroup(GraphVisualization.GRAPH_NODES_PROBLEMATIC_FIT);
		mode = Mode.NON_LIVE;
		changesMade = false;
		bfi = new BreadthFirstIterator();
		display.add(this, 0);
		initGui();
	}

	private void initGui() {
		setBounds(100, 100, 103, 27);
		setLayout(new BorderLayout());

		leftButton = new JButton(new LeftAction("", ImageHelper.loadImage("left-arrow-navigation.png"),
				"Navigation left", KeyEvent.VK_LEFT));
		leftButton.setPreferredSize(new Dimension(22, 9));
		add(leftButton, BorderLayout.WEST);

		rightButton = new JButton(new RightAction("", ImageHelper.loadImage("right-arrow-navigation.png"),
				"Navigation right", KeyEvent.VK_RIGHT));
		rightButton.setPreferredSize(new Dimension(22, 9));
		add(rightButton, BorderLayout.EAST);

		JPanel panel = new JPanel();
		final Point initialClick = new Point();
		panel.addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				initialClick.setLocation(e.getPoint());
				getComponentAt(initialClick);
			}
		});

		panel.addMouseMotionListener(new MouseMotionAdapter() {
			@Override
			public void mouseDragged(MouseEvent e) {

				// get location of Window
				int thisX = SoundnessInteractionGUI.this.getLocation().x;
				int thisY = SoundnessInteractionGUI.this.getLocation().y;

				// Determine how much the mouse moved since the initial click
				int xMoved = (thisX + e.getX()) - (thisX + initialClick.x);
				int yMoved = (thisY + e.getY()) - (thisY + initialClick.y);

				// Move window to this position
				int X = thisX + xMoved;
				int Y = thisY + yMoved;
				SoundnessInteractionGUI.this.setLocation(X, Y);
			}
		});
		add(panel, BorderLayout.CENTER);

		label = new JLabel("-/-");
		panel.add(label);
		label.setHorizontalAlignment(SwingConstants.CENTER);
		setVisible(false);
	}

	private void updateView() {
		if (current >= 0) {
			controller.setStartingNodeDecoratorEnabled(false);
			clearPreviousState();

			List<String> marking;
			String transitionKey;

			switch (mode) {
				case NON_LIVE :
					marking = controller.getSoundnessInformation().getNonLiveSequenceMarking(current);
					transitionKey = controller.getSoundnessInformation().getNonLiveSequenceTransition(current);
					break;
				default :
					marking = controller.getSoundnessInformation().getUnboundedSequenceMarking(current);
					transitionKey = controller.getSoundnessInformation().getUnboundedSequenceTransition(current);
					break;
			}

			VisualItem visualTransition = vis
					.getVisualItem(GraphVisualization.GRAPH_NODES_GROUP,
							(Tuple) vis.getGroup(GraphVisualization.GRAPH_NODES_GROUP)
									.tuples(ExpressionParser
											.predicate("[" + PNConstants.NODE_ID_KEY + "] = '" + transitionKey + "'"))
					.next());
			visualTransition.setHighlighted(true);
			focusGroupHighlight.addTuple(visualTransition);
			focusGroupFit.addTuple(visualTransition);

			for (String mark : marking) {
				VisualItem visualPlace = vis
						.getVisualItem(GraphVisualization.GRAPH_NODES_GROUP,
								(Tuple) vis.getGroup(GraphVisualization.GRAPH_NODES_GROUP)
										.tuples(ExpressionParser
												.predicate("[" + PNConstants.NODE_ID_KEY + "] = '" + mark + "'"))
						.next());
				visualPlace.setHighlighted(true);
				visualPlace.setInt(PNConstants.NODE_TOKEN_COUNT_KEY, 1);
				focusGroupHighlight.addTuple(visualPlace);
				focusGroupFit.addTuple(visualPlace);
			}
			
			bfi.init(focusGroupHighlight.tuples(), Integer.MAX_VALUE, Constants.NODE_TRAVERSAL);
			while (bfi.hasNext()) {
				Node n = (Node) bfi.next();
				n.setInt(PNConstants.NODE_DEPTH_KEY, bfi.getDepth(n));
				if (bfi.getDepth(n) == 1 && !focusGroupFit.containsTuple(n)) {
					focusGroupFit.addTuple(n);
				}
			}
		} else {
			clearPreviousState();
			controller.setStartingNodeDecoratorEnabled(true);
			List<String> keys;
			switch (mode) {
				case NON_LIVE :
					keys = controller.getSoundnessInformation().getNonLiveTransitions();
					break;
				case UNBOUNDED :
					keys = controller.getSoundnessInformation().getUnboundedPlaces();
					break;
				default :
					keys = controller.getSoundnessInformation().getDeadTransitions();
					break;
			}
			for (String key : keys) {
				VisualItem visualPlace = vis
						.getVisualItem(GraphVisualization.GRAPH_NODES_GROUP,
								(Tuple) vis.getGroup(GraphVisualization.GRAPH_NODES_GROUP)
										.tuples(ExpressionParser
												.predicate("[" + PNConstants.NODE_ID_KEY + "] = '" + key + "'"))
						.next());
				visualPlace.setHighlighted(true);
				focusGroupHighlight.addTuple(visualPlace);
				focusGroupFit.addTuple(visualPlace);
			}
		}
		changesMade = true;
		
		if (maxDepth == -1) {
			label.setText("-/-");
		} else {
			label.setText((current + 1) + "/" + (maxDepth + 1));
		}

		runActions();
	}

	private void runActions() {
		vis.cancel(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
		vis.run(PNConstants.VIS_DECORATORS_ACTIONS_NAME);
		vis.run(PNConstants.VIS_SOUNDNESS_COLOR_ACTIONS_NAME);
		vis.run(PNConstants.VIS_STROKE_COLOR_ACTIONS_NAME);
		vis.run(PNConstants.VIS_FILL_EDGES_ACTIONS_NAME);
		vis.run(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);

		if (focusGroupFit.getTupleCount() > 0) {
			vis.run(PNConstants.VIS_SOUNDNESS_PROBLEMATIC_FIT_ACTIONS_NAME);
		} else {
			vis.run(PNConstants.VIS_LAYOUT_FIT_BOUNDS_ANIMATED_ACTIONS_NAME);
		}
	}

	private void clearPreviousState() {
		focusGroupHighlight.clear();
		focusGroupFit.clear();
		vis.setValue(GraphVisualization.GRAPH_NODES_GROUP, null, VisualItem.HIGHLIGHT, false);
		vis.setValue(GraphVisualization.GRAPH_NODES_GROUP, null, PNConstants.NODE_TOKEN_COUNT_KEY, 0);
		vis.setValue(GraphVisualization.GRAPH_NODES_GROUP, null, PNConstants.NODE_DEPTH_KEY, 0);
	}

	public void setEnabledController(boolean enabled) {
		if (enabled) {
			updateView();
		} else {
			if (changesMade) {
				clearPreviousState();
				vis.run(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
				vis.run(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
				vis.run(PNConstants.VIS_STROKE_COLOR_ACTIONS_NAME);
				vis.run(PNConstants.VIS_FILL_EDGES_ACTIONS_NAME);
				vis.run(PNConstants.VIS_LAYOUT_FIT_BOUNDS_ANIMATED_ACTIONS_NAME);
				current = -1;
				changesMade = false;
			}
			controller.setStartingNodeDecoratorEnabled(true);
		}
		setVisible(enabled);
	}

	public void setNonLiveMode() {
		mode = Mode.NON_LIVE;
		current = -1;
		maxDepth = controller.getSoundnessInformation().getNonLiveSequenceSize() - 1;
		clearPreviousState();
	}

	public void setDeadMode() {
		mode = Mode.DEAD;
		current = -1;
		maxDepth = -1;
		clearPreviousState();
	}

	public void setUnboundedMode() {
		mode = Mode.UNBOUNDED;
		current = -1;
		maxDepth = controller.getSoundnessInformation().getUnboundedSequenceSize() - 1;
		clearPreviousState();
	}

	private class LeftAction extends AbstractAction {
		public LeftAction(String text, ImageIcon icon, String desc, Integer mnemonic) {
			super(text, icon);
			putValue(SHORT_DESCRIPTION, desc);
			putValue(MNEMONIC_KEY, mnemonic);
		}

		public void actionPerformed(ActionEvent e) {
			if (!vis.getDisplay(0).isTranformInProgress()) {
				current = Math.max(-1, current - 1);
				updateView();
			}
		}
	}

	private class RightAction extends AbstractAction {
		public RightAction(String text, ImageIcon icon, String desc, Integer mnemonic) {
			super(text, icon);
			putValue(SHORT_DESCRIPTION, desc);
			putValue(MNEMONIC_KEY, mnemonic);
		}

		public void actionPerformed(ActionEvent e) {
			if (!vis.getDisplay(0).isTranformInProgress()) {
				current = Math.min(current + 1, maxDepth);
				updateView();
			}
		}
	}
}
