package org.processmining.humbertogarciacaballero.models.gui.components;

import java.awt.Color;
import java.awt.Dimension;
import java.util.Map;
import java.util.Set;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.LineBorder;
import javax.swing.border.TitledBorder;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.prabhakardixit.processanalytics.ProcessAnalyticsViz;

import com.google.common.collect.Maps;
import com.jgoodies.forms.layout.ColumnSpec;
import com.jgoodies.forms.layout.FormLayout;
import com.jgoodies.forms.layout.FormSpecs;
import com.jgoodies.forms.layout.RowSpec;

import prefuse.visual.NodeItem;

public class TransitionStatisticsInformationPopUp extends JPanel {

	private NodeItem transition;
	private JLabel lblAbsolute;
	private JLabel lblCase;
	private JLabel lblAvgPerCase;
	private JLabel labelAbsoluteValue;
	private JLabel labelCaseValue;
	private JLabel labelAvgPerCaseValue;

	/**
	 * Create the panel.
	 */
	public TransitionStatisticsInformationPopUp() {
		setSize(new Dimension(400, 300));
		setLayout(new FormLayout(new ColumnSpec[] {
				FormSpecs.RELATED_GAP_COLSPEC,
				FormSpecs.DEFAULT_COLSPEC,
				FormSpecs.RELATED_GAP_COLSPEC,
				FormSpecs.DEFAULT_COLSPEC,},
			new RowSpec[] {
				FormSpecs.RELATED_GAP_ROWSPEC,
				FormSpecs.DEFAULT_ROWSPEC,
				FormSpecs.RELATED_GAP_ROWSPEC,
				FormSpecs.DEFAULT_ROWSPEC,
				FormSpecs.RELATED_GAP_ROWSPEC,
				FormSpecs.DEFAULT_ROWSPEC,}));
		
		lblAbsolute = new JLabel("Absolute:");
		add(lblAbsolute, "2, 2, right, default");
		
		labelAbsoluteValue = new JLabel("");
		add(labelAbsoluteValue, "4, 2, right, default");
		
		lblCase = new JLabel("Case:");
		add(lblCase, "2, 4, right, default");
		
		labelCaseValue = new JLabel("");
		add(labelCaseValue, "4, 4, right, default");
		
		lblAvgPerCase = new JLabel("Avg. per case:");
		add(lblAvgPerCase, "2, 6, right, default");
		
		labelAvgPerCaseValue = new JLabel("");
		add(labelAvgPerCaseValue, "4, 6, right, default");
	}

	public void populatePopUp(NodeItem transition) {
		clear();
		PNetVisualizerController controller = PNetVisualizerController.getInstance();

		// Look for the place object
		Transition transitionObject = null;
		for (PetrinetNode node : controller.getProcessAnalytics().getPetrinet().getNodes()) {
			if (node.getLocalID().localNodeID().toString().equals(transition.getString(PNConstants.NODE_ID_KEY))) {
				transitionObject = (Transition) node;
			}
		}

		String text;
		switch (ProcessAnalyticsViz.mode) {
			case 0:
				text = "Sync-model";
				break;
			case 1:
				text = "Frequency";
				break;
			default:
				text = "Performance";
				break;
		}

		setBorder(new TitledBorder(new LineBorder(Color.BLACK), text));
		
//		labelAbsoluteValue.setText(ProcessAnalyticsViz.current);
		
		// Store temporally
		this.transition = transition;

		invalidate();
		repaint();
	}

	private Map<String, Float> calculateSingleFrequency(Place place) {
		Map<String, Float> map = Maps.newHashMap();

		for (Set<Place> set : ProcessAnalyticsViz.logMovesFrequencyCurrent.keySet()) {
			if (set.contains(place) && set.size() == 1) {
				Map<String, Float> aux = ProcessAnalyticsViz.logMovesFrequencyCurrent.get(set);
				for (String activity : aux.keySet()) {
					if (map.containsKey(activity)) {
						map.put(activity, map.get(activity) + aux.get(activity));
					} else {
						map.put(activity, aux.get(activity));
					}
				}
			}
		}

		return map;
	}

	private Map<String, Float> calculateSeveralFrequency(Place place) {
		Map<String, Float> map = Maps.newHashMap();

		for (Set<Place> set : ProcessAnalyticsViz.logMovesFrequencyCurrent.keySet()) {
			if (set.contains(place) && set.size() > 1) {
				Map<String, Float> aux = ProcessAnalyticsViz.logMovesFrequencyCurrent.get(set);
				for (String activity : aux.keySet()) {
					if (map.containsKey(activity)) {
						map.put(activity, map.get(activity) + aux.get(activity));
					} else {
						map.put(activity, aux.get(activity));
					}
				}
			}
		}

		return map;
	}

	public NodeItem getTransition() {
		return transition;
	}

	public void clear() {
		labelAbsoluteValue.setText("");
		labelCaseValue.setText("");
		labelAvgPerCaseValue.setText("");
		transition = null;
	}
}
