package org.processmining.humbertogarciacaballero.models.gui.controllers;

import java.awt.event.MouseEvent;
import java.util.Iterator;

import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.Visualization;
import prefuse.controls.ControlAdapter;
import prefuse.data.Edge;
import prefuse.data.expression.Predicate;
import prefuse.util.ColorLib;
import prefuse.visual.AggregateItem;
import prefuse.visual.DecoratorItem;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;
import prefuse.visual.expression.InGroupPredicate;

public class LayerSimplifierControl extends ControlAdapter {

    private int currentLayer;
    private Predicate decoratorGroup;

    public LayerSimplifierControl(String group) {
        decoratorGroup = new InGroupPredicate(group);
    }

    /**
     * @see prefuse.controls.Control#itemEntered(prefuse.visual.VisualItem,
     *      java.awt.event.MouseEvent)
     */
    public void itemEntered(VisualItem item, MouseEvent e) {
        if (decoratorGroup.getBoolean(item)) {
            // Look for the nodes
            int layer = item.getInt(PNConstants.NODE_LAYER_KEY);
            int maxLayer = item.getInt(PNConstants.NODE_MAX_LAYER_KEY);

            item.set(VisualItem.STROKECOLOR, ColorLib.rgba(181, 16, 16, 150));

            toggleHihglightEdges((AggregateItem) ((DecoratorItem) item).getDecoratedItem(), true);
            runAnimations(item.getVisualization());
            currentLayer = layer == maxLayer ? layer - 1 : layer;
        }
    }

    /**
     * @see prefuse.controls.Control#itemExited(prefuse.visual.VisualItem,
     *      java.awt.event.MouseEvent)
     */
    public void itemExited(VisualItem item, MouseEvent e) {
        if (decoratorGroup.getBoolean(item)) {
            item.set(VisualItem.STROKECOLOR, ColorLib.rgba(255, 255, 255, 0));

            toggleHihglightEdges((AggregateItem) ((DecoratorItem) item).getDecoratedItem(), false);
            runAnimations(item.getVisualization());
        }
    }

    private void toggleHihglightEdges(AggregateItem item, boolean highlight) {
        int layer = item.getInt(PNConstants.NODE_LAYER_KEY);
        Iterator<NodeItem> iter = item.items();
        while (iter.hasNext()) {
            NodeItem node = iter.next();
            if (node.getInt(PNConstants.NODE_LAYER_KEY) == layer) {
                Iterator<EdgeItem> edges = node.edges();

                while (edges.hasNext()) {
                    EdgeItem edge = edges.next();
                    NodeItem adjacent = edge.getAdjacentItem(node);

                    if (adjacent.getInt(PNConstants.NODE_LAYER_KEY) >= layer) {
                        edge.setHighlighted(highlight);
                    }
                }
            }
        }
    }

    private void runAnimations(Visualization vis) {
        vis.run(PNConstants.VIS_FILL_EDGES_ACTIONS_NAME);
        vis.run(PNConstants.VIS_STROKE_EDGES_ACTIONS_NAME);
        vis.run(PNConstants.VIS_REPAINT_ACTIONS_NAME);
    }

    /**
     * @see prefuse.controls.Control#itemClicked(prefuse.visual.VisualItem,
     *      java.awt.event.MouseEvent)
     */
    @SuppressWarnings("unchecked")
    public void itemClicked(VisualItem item, MouseEvent e) {
        if (decoratorGroup.getBoolean(item)) {
            boolean simplify = checkAllSimplified(((AggregateItem) ((DecoratorItem) item).getDecoratedItem()));
            boolean runAnimations = false;
            Iterator<VisualItem> iter = ((AggregateItem) ((DecoratorItem) item).getDecoratedItem()).items();

            while (iter.hasNext()) {
                VisualItem vi = iter.next();

                if (vi instanceof NodeItem) {
                    NodeItem node = (NodeItem) vi;

                    if (node.getInt(PNConstants.NODE_LAYER_KEY) == currentLayer) {
                        Iterator<Edge> iter2 = node.edges();

                        while (iter2.hasNext()) {
                            Edge edge = iter2.next();

                            if (edge.getAdjacentNode(node).getInt(PNConstants.NODE_LAYER_KEY) > currentLayer) {
                                edge.setBoolean(PNConstants.EDGE_SIMPLIFICABLE, !simplify);
                                edge.setBoolean(PNConstants.EDGE_FORCE_SIMPLIFICATION, !simplify);
                                if (!simplify)
                                    edge.setBoolean(PNConstants.SIMPLIFIED_KEY, !simplify);
                                runAnimations = true;
                            }
                        }
                    }
                }
                item.setBoolean(PNConstants.SIMPLIFIED_KEY, !simplify);
            }

            if (runAnimations) {
                item.getVisualization().cancel(PNConstants.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
                item.getVisualization().run(PNConstants.VIS_LAYOUT_NO_PAINT_ACTIONS_NAME);
                item.getVisualization().run(PNConstants.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
            }

        }
    }

    private boolean checkAllSimplified(AggregateItem item) {
        boolean allSimplified = true;
        Iterator<VisualItem> iter = item.items();

        while (iter.hasNext()) {
            VisualItem vi = iter.next();
            if (vi instanceof NodeItem) {
                NodeItem node = (NodeItem) vi;
                if (node.getInt(PNConstants.NODE_LAYER_KEY) == currentLayer) {
                    Iterator<Edge> iter2 = node.edges();
                    while (iter2.hasNext()) {
                        Edge edge = iter2.next();
                        if (edge.getAdjacentNode(node).getInt(PNConstants.NODE_LAYER_KEY) > currentLayer) {
                            allSimplified &= edge.getBoolean(PNConstants.SIMPLIFIED_KEY);
                        }
                    }
                }
            }
        }

        return allSimplified;
    }
}
