package org.processmining.humbertogarciacaballero.models.gui.controllers;

import java.util.Iterator;

import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.controls.NeighborHighlightControl;
import prefuse.data.tuple.TupleSet;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class NodeHighlightControl extends NeighborHighlightControl {

	private TupleSet incomingNeighbors;

	private TupleSet outgoingNeighbors;

	public NodeHighlightControl(TupleSet incomingNeighbors, TupleSet outgoingNeighbors) {
		this.incomingNeighbors = incomingNeighbors;
		this.outgoingNeighbors = outgoingNeighbors;
	}

	/**
	 * Set the highlighted state of the neighbors of a node.
	 * 
	 * @param n
	 *            the node under consideration
	 * @param state
	 *            the highlighting state to apply to neighbors
	 */
	@SuppressWarnings("unchecked")
	protected void setNeighborHighlight(NodeItem n, boolean state) {
		Iterator<EdgeItem> iter = n.edges();
		while (iter.hasNext()) {
			EdgeItem eitem = (EdgeItem) iter.next();
			NodeItem nitem = eitem.getAdjacentItem(n);
			if (eitem.isVisible() || isHighlightWithInvisibleEdge()) {
				eitem.setHighlighted(state);
				nitem.setHighlighted(state);

				if (state) {
					if (eitem.getSourceItem().equals(nitem)) {
						incomingNeighbors.addTuple(nitem);
					} else {
						outgoingNeighbors.addTuple(nitem);
					}
				} else {
					if (eitem.getSourceItem().equals(nitem)) {
						incomingNeighbors.removeTuple(nitem);
					} else {
						outgoingNeighbors.removeTuple(nitem);
					}
				}
			}
		}

		n.setHighlighted(state);

		n.getVisualization().cancel(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
		n.getVisualization().run(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
		n.getVisualization().run(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
	}

	public void itemKeyPressed(VisualItem item, java.awt.event.KeyEvent e) {

	}
}
