package org.processmining.humbertogarciacaballero.models.gui.controllers;

import java.awt.Cursor;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Iterator;

import javax.swing.SwingUtilities;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.gui.GraphVisualization;
import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.Display;
import prefuse.Visualization;
import prefuse.controls.ControlAdapter;
import prefuse.data.Tuple;
import prefuse.data.tuple.TupleSet;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class SelectNodeControl extends ControlAdapter {

	private boolean transitionSelection;
	private int downX1, downY1;
	private VisualItem rubberBand;
	private Point2D screenPoint = new Point2D.Float();
	private Point2D absPoint = new Point2D.Float();
	private Rectangle2D rect = new Rectangle2D.Float();
	private PNetVisualizerController controller;

	public SelectNodeControl(VisualItem rubberBand) {
		super();
		this.rubberBand = rubberBand;
		this.controller = PNetVisualizerController.getInstance();
	}

	public void mousePressed(MouseEvent e) {
		if (!SwingUtilities.isLeftMouseButton(e))
			return;

		Display d = (Display) e.getComponent();
		Visualization vis = d.getVisualization();
		TupleSet focus = vis.getFocusGroup(GraphVisualization.SELECTED_NODES);
		if (!e.isShiftDown()) {
			focus.clear();
		}

		float[] bandRect = (float[]) rubberBand.get(VisualItem.POLYGON);
		bandRect[0] = bandRect[1] = bandRect[2] = bandRect[3] = bandRect[4] = bandRect[5] = bandRect[6] = bandRect[7] = 0;

		screenPoint.setLocation(e.getX(), e.getY());
		d.getAbsoluteCoordinate(screenPoint, absPoint);
		downX1 = (int) absPoint.getX();
		downY1 = (int) absPoint.getY();
		rubberBand.setVisible(true);
	}

	@SuppressWarnings("unchecked")
	public void mouseDragged(MouseEvent e) {
		if (!SwingUtilities.isLeftMouseButton(e))
			return;

		Display d = (Display) e.getComponent();
		screenPoint.setLocation(e.getX(), e.getY());
		d.getAbsoluteCoordinate(screenPoint, absPoint);
		int x1 = downX1;
		int y1 = downY1;
		int x2 = (int) absPoint.getX();
		int y2 = (int) absPoint.getY();

		float[] bandRect = (float[]) rubberBand.get(VisualItem.POLYGON);
		bandRect[0] = x1;
		bandRect[1] = y1;
		bandRect[2] = x2;
		bandRect[3] = y1;
		bandRect[4] = x2;
		bandRect[5] = y2;
		bandRect[6] = x1;
		bandRect[7] = y2;

		if (x2 < x1) {
			int temp = x2;
			x2 = x1;
			x1 = temp;
		}
		if (y2 < y1) {
			int temp = y2;
			y2 = y1;
			y1 = temp;
		}
		rect.setRect(x1, y1, x2 - x1, y2 - y1);

		Visualization vis = d.getVisualization();
		TupleSet focus = vis.getFocusGroup(GraphVisualization.SELECTED_NODES);

		if (!e.isShiftDown()) {
			focus.clear();
		}

		Tuple[] selectedItems = new Tuple[vis.getGroup(GraphVisualization.GRAPH_NODES_GROUP).getTupleCount()];
		Iterator<VisualItem> it = vis.getGroup(GraphVisualization.GRAPH_NODES_GROUP).tuples();

		int i = 0;
		while (it.hasNext()) {
			VisualItem item = it.next();
			if (item.isVisible() && item.getBounds().intersects(rect)) {
				selectedItems[i++] = item;
			}
		}

		Tuple[] properlySizedSelectedItems = new Tuple[i];
		System.arraycopy(selectedItems, 0, properlySizedSelectedItems, 0, i);
		for (int j = 0; j < properlySizedSelectedItems.length; j++) {
			Tuple tuple = properlySizedSelectedItems[j];
			if (((NodeType.valueOf(tuple.getString(PNConstants.NODE_TYPE_KEY)) == NodeType.TRANSITION
					&& transitionSelection)
					|| ((NodeType.valueOf(tuple.getString(PNConstants.NODE_TYPE_KEY)) == NodeType.PLACE
							&& !transitionSelection))) && !tuple.getBoolean(PNConstants.NODE_SELECTED_TRANSITION)) {
				focus.addTuple(tuple);
				// TODO: REFRESH STATISTICS
			}
		}

		rubberBand.setValidated(false);
		d.repaint();

	}

	public void mouseReleased(MouseEvent e) {
		if (!SwingUtilities.isLeftMouseButton(e))
			return;
		rubberBand.setVisible(false);
		Display d = (Display) e.getComponent();

		d.setHighQuality(true);
		d.getVisualization().repaint();
		//		removeSelectedNodes(d.getVisualization(), d);
	}

	/**
	 * @see prefuse.controls.Control#itemEntered(prefuse.visual.VisualItem,
	 *      java.awt.event.MouseEvent)
	 */
	@Override
	public void itemEntered(VisualItem item, MouseEvent e) {
		if (item instanceof NodeItem) {
			Display d = (Display) e.getSource();
			d.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
		}
	}

	/**
	 * @see prefuse.controls.Control#itemExited(prefuse.visual.VisualItem,
	 *      java.awt.event.MouseEvent)
	 */
	@Override
	public void itemExited(VisualItem item, MouseEvent e) {
		if (item instanceof NodeItem) {
			Display d = (Display) e.getSource();
			d.setCursor(Cursor.getDefaultCursor());
		}
	}

	@Override
	public void itemClicked(VisualItem item, MouseEvent e) {
		if (!SwingUtilities.isLeftMouseButton(e))
			return;
		if (item instanceof NodeItem) {
			Display d = (Display) e.getComponent();
			Visualization vis = d.getVisualization();
			if (((NodeType.valueOf(item.getString(PNConstants.NODE_TYPE_KEY)) == NodeType.TRANSITION
					&& transitionSelection)
					|| ((NodeType.valueOf(item.getString(PNConstants.NODE_TYPE_KEY)) == NodeType.PLACE
							&& !transitionSelection))) && !item.getBoolean(PNConstants.NODE_SELECTED_TRANSITION)) {
				vis.getFocusGroup(GraphVisualization.SELECTED_NODES).addTuple(item);
				d.repaint();

				// TODO: REFRESH STATISTICS
			}
			//			removeSelectedNodes(vis, d);
		}
	}

	//	/**
	//	 * @see prefuse.controls.Control#keyTyped(java.awt.event.KeyEvent)
	//	 */
	//	public void keyReleased(KeyEvent e) {
	//		if (e.getKeyCode() == KeyEvent.VK_DELETE) {
	//			Display d = (Display) e.getComponent();
	//			Visualization vis = d.getVisualization();
	//			removeSelectedNodes(vis, d);
	//		}
	//	}

	//	@SuppressWarnings("unchecked")
	//	private void removeSelectedNodes(Visualization vis, Display display) {
	//		DefaultTupleSet ts = (DefaultTupleSet) vis.getFocusGroup(GraphVisualization.SELECTED_NODES);
	//		if (ts.getTupleCount() > 0) {
	//			int choice = showConfirmationDialog(display);
	//			if (choice == JOptionPane.YES_OPTION) {
	//				Iterator<NodeItem> iter = ts.tuples();
	//				while (iter.hasNext()) {
	//					controller.hideNode(iter.next());
	//				}
	//			}
	//			ts.clear();
	//		}
	//	}

	//	private int showConfirmationDialog(Display display) {
	//		return JOptionPane.showConfirmDialog(display, "Do you want to remove the selected node(s)?", "Detele node(s)?",
	//				JOptionPane.YES_NO_OPTION);
	//	}

	public void setTransitionSelection(boolean b) {
		transitionSelection = b;
	}
}
