package org.processmining.humbertogarciacaballero.models.gui.render;

import static org.processmining.humbertogarciacaballero.models.util.Utils.calculateOffsetPortPoint;
import static org.processmining.humbertogarciacaballero.models.util.Utils.calculatePortPoint;

import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.LinkedList;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;

import prefuse.render.EdgeRenderer;
import prefuse.util.GraphicsLib;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class PolylineEdgeRenderer extends EdgeRenderer {

	protected Path2D path;
	protected HierarchicalLayoutContext context;
	protected static final List<Point2D> EMPTY_LIST = new LinkedList<>();
	protected Point2D[] tempPoints = new Point2D.Double[] { new Point2D.Double(), new Point2D.Double() };

	public PolylineEdgeRenderer() {
		super();
		path = new Path2D.Double();
		context = HierarchicalLayoutContext.getInstance();
	}

	@SuppressWarnings("unchecked")
	@Override
	protected Shape getRawShape(VisualItem item) {
		List<Point2D> bends = item.get(PNConstants.EDGE_BENDS_KEY) != null
				? new LinkedList<>((List<Point2D>) item.get(PNConstants.EDGE_BENDS_KEY)) : EMPTY_LIST;
		EdgeItem edgeItem = (EdgeItem) item;
		NodeItem sourceItem = edgeItem.getSourceItem();
		NodeItem targetItem = edgeItem.getTargetItem();
		int fromLayer = sourceItem.getInt(PNConstants.NODE_LAYER_KEY),
				toLayer = targetItem.getInt(PNConstants.NODE_LAYER_KEY);

		m_tmpPoints[0].setLocation(calculateOffsetPortPoint(edgeItem, targetItem, false, fromLayer >= toLayer));
		m_tmpPoints[1].setLocation(calculatePortPoint(edgeItem, targetItem, false));

		m_curWidth = (float) (m_width * getLineWidth(item));

		Path2D arrowPath = (Path2D) createArrowHead(m_tmpPoints[0], m_tmpPoints[1], true,
				NodeType.TRANSITION == NodeType.valueOf(targetItem.getString(PNConstants.NODE_TYPE_KEY)),
				targetItem.getBounds());

		if (edgeItem.getBoolean(PNConstants.EDGE_BIDIRECTIONAL)) {
			tempPoints[0].setLocation(calculateOffsetPortPoint(edgeItem, sourceItem, true, fromLayer <= toLayer));
			tempPoints[1].setLocation(calculatePortPoint(edgeItem, sourceItem, true));
			arrowPath.append(createArrowHead(tempPoints[0], tempPoints[1], true,
					NodeType.TRANSITION == NodeType.valueOf(sourceItem.getString(PNConstants.NODE_TYPE_KEY)),
					sourceItem.getBounds()), false);
			m_isctPoints[0].setLocation(tempPoints[0]);
			tempPoints[0].setLocation(tempPoints[1]);
			tempPoints[1].setLocation(m_isctPoints[0]);
		} else {
			if (NodeType.PLACE == NodeType.valueOf(sourceItem.getString(PNConstants.NODE_TYPE_KEY))) {
				if (intersectLineCircle(calculateOffsetPortPoint(edgeItem, sourceItem, true, fromLayer <= toLayer),
						calculatePortPoint(edgeItem, sourceItem, true),
						new Point2D.Double(sourceItem.getBounds().getCenterX(), sourceItem.getBounds().getCenterY()),
						sourceItem.getBounds().getWidth() / 2, m_isctPoints) > 0) {
					tempPoints[0].setLocation(m_isctPoints[0]);
				} else {
					tempPoints[0].setLocation(calculatePortPoint(edgeItem, sourceItem, true));
				}
			} else {
				if (GraphicsLib.intersectLineRectangle(calculatePortPoint(edgeItem, sourceItem, true),
						calculateOffsetPortPoint(edgeItem, sourceItem, true, fromLayer <= toLayer),
						sourceItem.getBounds(), m_isctPoints) > 0)
					tempPoints[0].setLocation(m_isctPoints[0]);
				else
					tempPoints[0].setLocation(calculatePortPoint(edgeItem, sourceItem, true));
			}
			tempPoints[1].setLocation(calculateOffsetPortPoint(edgeItem, sourceItem, true, fromLayer <= toLayer));
		}

		m_curArrow = arrowPath;

		path.reset();
		path.moveTo(tempPoints[0].getX(), tempPoints[0].getY());
		path.lineTo(tempPoints[1].getX(), tempPoints[1].getY());

		for (int i = 0; i < bends.size(); i++) {
			Point2D point = bends.get(i);
			path.lineTo(point.getX(), point.getY());
		}

		path.lineTo(m_tmpPoints[0].getX(), m_tmpPoints[0].getY());
		path.lineTo(m_tmpPoints[1].getX(), m_tmpPoints[1].getY());

		return path;
	}

	protected int intersectLineCircle(Point2D pointA, Point2D pointB, Point2D center, double radius, Point2D[] pts) {
		double baX = pointB.getX() - pointA.getX();
		double baY = pointB.getY() - pointA.getY();
		double caX = center.getX() - pointA.getX();
		double caY = center.getY() - pointA.getY();

		double a = baX * baX + baY * baY;
		double bBy2 = baX * caX + baY * caY;
		double c = caX * caX + caY * caY - radius * radius;

		double pBy2 = bBy2 / a;
		double q = c / a;

		double disc = pBy2 * pBy2 - q;
		if (disc < 0) {
			return 0;
		}

		double tmpSqrt = Math.sqrt(disc);
		double abScalingFactor1 = -pBy2 + tmpSqrt;
		double abScalingFactor2 = -pBy2 - tmpSqrt;

		pts[0] = new Point2D.Double(pointA.getX() - baX * abScalingFactor1, pointA.getY() - baY * abScalingFactor1);
		if (disc == 0) {
			return 1;
		}
		pts[1] = new Point2D.Double(pointA.getX() - baX * abScalingFactor2, pointA.getY() - baY * abScalingFactor2);
		return 2;
	}

	@Override
	protected double getLineWidth(VisualItem item) {
		return item.getStroke().getLineWidth();
	}

	/**
	 * @see prefuse.render.Renderer#locatePoint(java.awt.geom.Point2D,
	 *      prefuse.visual.VisualItem)
	 */
	@Override
	public boolean locatePoint(Point2D p, VisualItem item) {
		Shape s = getShape(item);
		if (s == null) {
			return false;
		} else {
			double width = Math.max(2.0, getLineWidth(item));
			double halfWidth = width / 2.0;
			return s.intersects(p.getX() - halfWidth, p.getY() - halfWidth, width, width);
		}
	}

	protected Shape createArrowHead(Point2D start, Point2D end, boolean computeIntersection,
			boolean rectangleIntersection, Rectangle2D intersectionBounds) {

		Point2D endAux = end;
		if (computeIntersection) {
			if (rectangleIntersection) {
				int i = GraphicsLib.intersectLineRectangle(start, end, intersectionBounds, m_isctPoints);
				if (i > 0)
					endAux = m_isctPoints[0];
			} else {
				int i = intersectLineCircle(start, end,
						new Point2D.Double(intersectionBounds.getCenterX(), intersectionBounds.getCenterY()),
						intersectionBounds.getWidth() / 2, m_isctPoints);
				if (i > 0)
					endAux = m_isctPoints[0];
			}
		}

		AffineTransform at = getArrowTrans(start, endAux, m_curWidth);
		Shape arrowShape = at.createTransformedShape(m_arrowHead);

		Point2D lineEnd = end;
		lineEnd.setLocation(0, -m_arrowHeight);
		at.transform(lineEnd, lineEnd);

		return arrowShape;
	}
}
