package org.processmining.humbertogarciacaballero.models.gui.render;

import static org.processmining.humbertogarciacaballero.models.util.Utils.calculateOffsetPortPoint;
import static org.processmining.humbertogarciacaballero.models.util.Utils.calculatePortPoint;

import java.awt.Shape;
import java.awt.geom.GeneralPath;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.util.Utils;

import prefuse.util.GraphicsLib;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class SimplifiedEdgeRenderer extends PolylineEdgeRenderer {

    private Path2D path;

    private boolean oriented;
    
    private Point2D[] auxPoints = {new Point2D.Double(), new Point2D.Double()};

    public SimplifiedEdgeRenderer() {
        super();
        path = new GeneralPath();
    }

    @Override
    protected Shape getRawShape(VisualItem item) {
        EdgeItem edgeItem = (EdgeItem) item;
        NodeItem sourceItem = edgeItem.getSourceItem();
        NodeItem targetItem = edgeItem.getTargetItem();
        int fromLayer = sourceItem.getInt(PNConstants.NODE_LAYER_KEY),
                toLayer = targetItem.getInt(PNConstants.NODE_LAYER_KEY);
        NodeType sourceType = NodeType.valueOf(sourceItem.getString(PNConstants.NODE_TYPE_KEY));
        NodeType targetType = NodeType.valueOf(targetItem.getString(PNConstants.NODE_TYPE_KEY));
        Rectangle2D intersectionBounds = new Rectangle2D.Float();
        intersectionBounds.setFrame(targetItem.getBounds());
        boolean placeToTransition;

        // Points
        Point2D p0, p1, p2, p3;
        if (oriented) {
            if (sourceType == NodeType.PLACE) {
                p0 = calculatePortPoint(edgeItem, sourceItem, true);

                double distance = context.getEdgeLandingOffset();
                switch (context.getLayoutOrientation()) {
                case TOP_TO_BOTTOM:
                case BOTTOM_TO_TOP:
                    distance += Utils.getHeight(sourceItem) / 2;
                    break;
                default:
                    distance += Utils.getWidth(sourceItem) / 2;
                    break;
                }
                p3 = calculatePortPoint(edgeItem, targetItem, false);
                p2 = calculateOffsetPortPoint(edgeItem, targetItem, false, fromLayer > toLayer);
                double fi = Math.atan2(p2.getY() - p0.getY(), p2.getX() - p0.getX());

                double x2 = p0.getX() + distance * Math.cos(fi);
                double y2 = p0.getY() + distance * Math.sin(fi);
                p1 = new Point2D.Double(x2, y2);
            } else {
                p0 = calculatePortPoint(edgeItem, sourceItem, true);
                p1 = calculateOffsetPortPoint(edgeItem, sourceItem, true, fromLayer <= toLayer);
                double distance = context.getEdgeLandingOffset();
                switch (context.getLayoutOrientation()) {
                case TOP_TO_BOTTOM:
                case BOTTOM_TO_TOP:
                    distance += Utils.getHeight(targetItem) / 2;
                    break;
                default:
                    distance += Utils.getWidth(targetItem) / 2;
                    break;
                }
                p3 = calculatePortPoint(edgeItem, targetItem, false);
                double fi = Math.atan2(p1.getY() - p3.getY(), p1.getX() - p3.getX());

                double x2 = p3.getX() + distance * Math.cos(fi);
                double y2 = p3.getY() + distance * Math.sin(fi);
                p2 = new Point2D.Double(x2, y2);
            }
        } else {
            p0 = calculatePortPoint(edgeItem, sourceItem, true);
            p1 = calculateOffsetPortPoint(edgeItem, sourceItem, true, fromLayer <= toLayer);
            p2 = calculateOffsetPortPoint(edgeItem, targetItem, false, fromLayer > toLayer);
            p3 = calculatePortPoint(edgeItem, targetItem, false);
        }
        m_tmpPoints[0].setLocation(p0);
        m_tmpPoints[1].setLocation(p1);
        tempPoints[0].setLocation(p2);
        tempPoints[1].setLocation(p3);

        if (sourceType == NodeType.PLACE && targetType == NodeType.TRANSITION) {
            placeToTransition = true;
        } else {
            placeToTransition = false;
        }

        m_curWidth = (float) (m_width * getLineWidth(item));

        /*---------------ARROW HEAD OPERATIONS-------------------*/
        /*-------CREATES TWO ARROWS FOR SIMPLIFIED EDGES---------*/
        // get starting and ending edge endpoints
        Path2D arrowPath = (Path2D) createArrowHead(m_tmpPoints[0], m_tmpPoints[1], false, placeToTransition,
                intersectionBounds);
        arrowPath.append(
                createArrowHead(tempPoints[0], tempPoints[1], true, placeToTransition, intersectionBounds),
                false);
        
        if (edgeItem.getBoolean(PNConstants.EDGE_BIDIRECTIONAL)) {
        	auxPoints[0].setLocation(m_tmpPoints[1]);
        	auxPoints[1].setLocation(m_tmpPoints[0]);
        	arrowPath.append(createArrowHead(auxPoints[0], auxPoints[1], true, !placeToTransition, sourceItem.getBounds()), false);
        	auxPoints[0].setLocation(auxPoints[1]);
        	m_tmpPoints[0].setLocation(tempPoints[1]);
        	arrowPath.append(createArrowHead(m_tmpPoints[0], tempPoints[0], false, placeToTransition, intersectionBounds), false);
        } else {
            if (placeToTransition) {
                if (intersectLineCircle(p1, p0,
                        new Point2D.Double(sourceItem.getBounds().getCenterX(), sourceItem.getBounds().getCenterY()),
                        sourceItem.getBounds().getWidth() / 2, m_isctPoints) > 0) {
                    auxPoints[0].setLocation(m_isctPoints[0]);
                } else {
                	auxPoints[0].setLocation(p0);
                }
            } else {
                if (GraphicsLib.intersectLineRectangle(p1, p0, sourceItem.getBounds(), m_isctPoints) > 0) {
                	auxPoints[0].setLocation(m_isctPoints[0]);
                } else {
                	auxPoints[0].setLocation(p0);
                }
            }
        }

        m_curArrow = arrowPath;

        // Construction of path
        path.reset();

        Point2D point = auxPoints[0];
        path.moveTo(point.getX(), point.getY());

        point = m_tmpPoints[1];
        path.lineTo(point.getX(), point.getY());

        point = tempPoints[0];
        path.moveTo(point.getX(), point.getY());

        point = tempPoints[1];
        path.lineTo(point.getX(), point.getY());

        return path;
    }

    public boolean isOrientedEnabled() {
        return oriented;
    }

    public void setOrientedEnabled(boolean oriented) {
        this.oriented = oriented;
    }
}
