package org.processmining.humbertogarciacaballero.models.model.commands;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.Visualization;
import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;

public class AddNodeCommand extends AbstractCommand {

    private PNetVisualizerController controller;
    private Visualization vis;
    private Graph graph;
    private Node newNode;
    private Node newNode2;
    private Node node;
    private boolean afterwards;

    private Node source;
    private Node target;

    public AddNodeCommand(Visualization vis, Graph graph, Node node, boolean afterwards) {
        this(vis, graph);

        this.node = node;
        this.afterwards = afterwards;
    }

    public AddNodeCommand(Visualization vis, Graph graph, Node source, Node target) {
        this(vis, graph);

        if (NodeType.valueOf(source.getString(PNConstants.NODE_TYPE_KEY)) != NodeType
                .valueOf(target.getString(PNConstants.NODE_TYPE_KEY)))
            throw new IllegalArgumentException("Nodes have to be of the same type.");

        this.source = source;
        this.target = target;
    }

    private AddNodeCommand(Visualization vis, Graph graph) {
        this.vis = vis;
        this.graph = graph;
        controller = PNetVisualizerController.getInstance();
    }

    @Override
    public void doIt() {
        if (source != null && target != null) {
            addNodeBetweenSourceAndTarget();
        } else if (afterwards) {
            addNodeAfterNode();
        } else {
            addNodeBeforeNode();
        }

        vis.run(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
        vis.run(PNConstants.VIS_SHAPE_ASSIGNMENT_ACTIONS_NAME);

        vis.run(PNConstants.VIS_LAYOUT_ACTIONS_NAME);
        finish();
    }

    private void addNodeBetweenSourceAndTarget() {
        newNode = graph.addNode();
        NodeType nodeType = NodeType.valueOf(source.getString(PNConstants.NODE_TYPE_KEY)).opposite();
        initializeNodeData(newNode, nodeType);
        graph.addEdge(source, newNode);
        graph.addEdge(newNode, target);
    }

    private void removeNodeBetweenSourceAndTarget() {
        graph.removeNode(newNode);
        newNode = null;
    }

    @SuppressWarnings("unchecked")
    private void addNodeAfterNode() {
        newNode = graph.addNode();
        NodeType nodeType = NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY)).opposite();
        initializeNodeData(newNode, nodeType);

        nodeType = nodeType.opposite();
        newNode2 = graph.addNode();
        initializeNodeData(newNode2, nodeType);

        graph.addEdge(newNode, newNode2);

        List<Edge> edgeList = new LinkedList<>();
        Iterator<Node> iter = node.outNeighbors();
        while (iter.hasNext()) {
            Node neighbor = iter.next();
            edgeList.add(graph.getEdge(node, neighbor));
            graph.addEdge(newNode2, neighbor);
        }

        for (Edge edge : edgeList) {
            graph.removeEdge(edge);
        }
        edgeList.clear();

        graph.addEdge(node, newNode);
    }

    @SuppressWarnings("unchecked")
    private void removeNodeAfterNode() {
        Iterator<Node> iter = newNode2.outNeighbors();
        while (iter.hasNext()) {
            graph.addEdge(node, iter.next());
        }
        graph.removeNode(newNode);
        graph.removeNode(newNode2);
        newNode = newNode2 = null;
    }

    @SuppressWarnings("unchecked")
    private void addNodeBeforeNode() {
        newNode = graph.addNode();
        NodeType nodeType = NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY)).opposite();
        initializeNodeData(newNode, nodeType);

        nodeType = nodeType.opposite();
        newNode2 = graph.addNode();
        initializeNodeData(newNode2, nodeType);

        graph.addEdge(newNode2, newNode);

        List<Edge> edgeList = new LinkedList<>();
        Iterator<Node> iter = node.inNeighbors();
        while (iter.hasNext()) {
            Node neighbor = iter.next();
            edgeList.add(graph.getEdge(neighbor, node));
            graph.addEdge(neighbor, newNode2);
        }

        for (Edge edge : edgeList) {
            graph.removeEdge(edge);
        }
        edgeList.clear();

        graph.addEdge(newNode, node);
    }
    
    private void initializeNodeData(Node node, NodeType nodeType) {
        newNode.setInt(PNConstants.NODE_WIDTH_KEY,
                nodeType == NodeType.PLACE ? controller.getPlacesSize() : controller.getTransitionsWidth());
        newNode.setInt(PNConstants.NODE_HEIGHT_KEY,
                nodeType == NodeType.PLACE ? controller.getPlacesSize() : controller.getTransitionsHeight());
        newNode.setString(PNConstants.NODE_TYPE_KEY, nodeType.toString());
        newNode.setString(PNConstants.NODE_LABEL_KEY, "New node");
    }

    @SuppressWarnings("unchecked")
    private void removeNodeBeforeNode() {
        Iterator<Node> iter = newNode2.inNeighbors();
        while (iter.hasNext()) {
            graph.addEdge(iter.next(), node);
        }
        graph.removeNode(newNode);
        graph.removeNode(newNode2);
        newNode = newNode2 = null;
    }

    @Override
    public void undoIt() {
        if (source != null && target != null) {
            removeNodeBetweenSourceAndTarget();
        } else if (afterwards) {
            removeNodeAfterNode();
        } else {
            removeNodeBeforeNode();
        }
        vis.run(PNConstants.VIS_LAYOUT_ACTIONS_NAME);
        finish();
    }

}
