package org.processmining.humbertogarciacaballero.models.model.commands;

import java.util.Iterator;

import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.Visualization;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;

public class HideNodeCommand extends AbstractCommand {

    private Visualization vis;
    private NodeItem node;

    public HideNodeCommand(NodeItem node) {
        this.node = node;
        this.vis = node.getVisualization();
    }

    @Override
    public void doIt() {
        toggleVisibility(false);
        executeActions();
        finish();
    }

    @Override
    public void undoIt() {
        toggleVisibility(true);
        executeActions();
        finish();
    }

    @SuppressWarnings("unchecked")
    private void toggleVisibility(boolean visible) {
        node.setVisible(visible);

        Iterator<EdgeItem> iter = node.inEdges();
        while (iter.hasNext()) {
            EdgeItem e = iter.next();
            NodeItem s = e.getSourceItem();
            NodeItem t = e.getTargetItem();
            e.setVisible(visible && s.isVisible() && t.isVisible());

            int outDegree = calculateDegree(s.outEdges());
            
            if (s.getOutDegree() == outDegree) {
                s.setInt(PNConstants.VISIBLE_OUT_DEGREE, -1);
            } else {
                s.setInt(PNConstants.VISIBLE_OUT_DEGREE, outDegree);
            }
        }

        iter = node.outEdges();
        while (iter.hasNext()) {
            EdgeItem e = iter.next();
            NodeItem s = e.getSourceItem();
            NodeItem t = e.getTargetItem();
            e.setVisible(visible && s.isVisible() && t.isVisible());

            int inDegree = calculateDegree(t.inEdges());
            
            if (t.getInDegree() == inDegree) {
                t.setInt(PNConstants.VISIBLE_IN_DEGREE, -1);
            } else {
                t.setInt(PNConstants.VISIBLE_IN_DEGREE, inDegree);
            }
        }
    }

    private int calculateDegree(Iterator<EdgeItem> iter) {
        int degree = 0;
        while (iter.hasNext()) {
            if (iter.next().isVisible())
                degree++;
        }
        return degree;
    }

    private void executeActions() {
        vis.run(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
        vis.run(PNConstants.VIS_DECORATORS_ACTIONS_NAME);
        vis.run(PNConstants.VIS_REPAINT_ACTIONS_NAME);
    }
}
