package org.processmining.humbertogarciacaballero.models.model.layout.stage.cycle;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.processmining.humbertogarciacaballero.models.util.GraphUtil;

import com.google.common.collect.Sets;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;


public class CycleDetector {

    // Prevents instantiation
    private CycleDetector() {

    }

    @SuppressWarnings("unchecked")
    public static void detectCycles(Graph graph, CycleVisitor visitor) {
        Graph graphCopy = GraphUtil.copySparseGraph(graph);
        LinkedList<Node> sl = new LinkedList<>();
        LinkedList<Node> sr = new LinkedList<>();

        while (graphCopy.getNodeCount() > 0) {
            Node node;
            while ((node = GraphUtil.getSink(graphCopy)) != null) {
                sr.addFirst(graph.getNodeFromKey(graphCopy.getKey(node.getRow())));
                graphCopy.removeNode(node);
            }

            while ((node = GraphUtil.getSource(graphCopy)) != null) {
                sl.addLast(graph.getNodeFromKey(graphCopy.getKey(node.getRow())));
                graphCopy.removeNode(node);
            }

            if (graphCopy.getNodeCount() > 0) {
                node = GraphUtil.getMaximumDifferenceNode(graphCopy);
                sl.addLast(graph.getNodeFromKey(graphCopy.getKey(node.getRow())));
                graphCopy.removeNode(node);
            }
        }

        List<Node> s = new LinkedList<>(sl);
        s.addAll(sr);

        for (int i = 0; i < s.size(); i++) {
            Node source = s.get(i);
            Set<Node> neighbours = Sets.newHashSet(source.outNeighbors());
            for (Node neighbour : neighbours) {
                if (i > s.indexOf(neighbour)) {
                    Edge edge = graph.getEdge(source, neighbour);
                    visitor.visitCycle(graph, source, neighbour, edge);
                }
            }
        }
    }
}
