package org.processmining.humbertogarciacaballero.models.model.layout.stage.emplacer;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.NodeParams;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayerContainer;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;

public class HeuristicEmplacer extends Emplacer {

	private int maxIterations = 8;

	private WeightedComparator comparator;

	public HeuristicEmplacer(HierarchicalLayoutContext context) {
		super(context);
		comparator = new WeightedComparator();
	}

	@Override
	public Map<Node, Float> calculateCoordinates(Graph graph) {
		Map<Node, Float> xcoords = initXCoords(graph);
		Map<Node, Float> xbest = new HashMap<>(xcoords);

		for (int i = 0; i < maxIterations; i++) {
			medianPos(i, xcoords);
			minEdge(i, xcoords);
			minNode(i, xcoords);
			minPath(graph, xcoords);
			packcut(xcoords);

			if (length(graph, xcoords) < length(graph, xbest)) {
				xbest = new HashMap<>(xcoords);
			}
		}

		return xbest;
	}

	@SuppressWarnings("unchecked")
    private float length(Graph graph, Map<Node, Float> xcoords) {
		float length = 0;

		// Objective function
		Iterator<Edge> iter = graph.edges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			length += getWeight(e) * Math.abs(xcoords.get(e.getTargetNode()) - xcoords.get(e.getSourceNode()));
		}

		return length;
	}

	private void packcut(Map<Node, Float> xcoords) {
		final Map<Node, Float> newXcoords = new HashMap<>(xcoords);
		List<Node> keys = new ArrayList<>(newXcoords.keySet());
		Collections.sort(keys, new Comparator<Node>() {
			@Override
			public int compare(Node o1, Node o2) {
				return Float.compare(newXcoords.get(o1), newXcoords.get(o2));
			}
		});

		for (int i = 0; i < keys.size(); i++) {
			Node u = keys.get(i);
			float currentX = newXcoords.get(u);
			boolean shiftPossible = true;

			for (int j = i; j < keys.size(); j++) {

			}
		}
	}

	private void minPath(Graph graph, Map<Node, Float> xcoords) {
		Iterator<Node> iter = graph.nodes();
		while (iter.hasNext()) {
			Node n = iter.next();
			if (((NodeParams) n.get(HierarchicalLayouter.NODE_PARAMS)).dummies.size() > 0) {
				for (List<Node> dummyNodes : ((NodeParams) n.get(HierarchicalLayouter.NODE_PARAMS)).dummies.values()) {

					int numDummies = dummyNodes.size();

					if (numDummies > 2) {
						float referenceX = xcoords.get(dummyNodes.get(0));
						boolean edgeStraight = true;
						int count = 0;

						for (int i = 1; i < numDummies; i++) {
							float x = xcoords.get(dummyNodes.get(i));

							if (x != referenceX) {
								edgeStraight = false;
								referenceX = x;
							} else {
								count++;
							}
						}

						if (!edgeStraight) {
							int upCount = 0;
							int downCount = 0;
							float upXPositions[] = new float[numDummies - 1];
							float downXPositions[] = new float[numDummies - 1];

							float currentX = xcoords.get(dummyNodes.get(0));

							for (int i = 0; i < numDummies - 1; i++) {
								float nextX = xcoords.get(dummyNodes.get(i + 1));

								if (currentX == nextX) {
									upXPositions[i] = currentX;
									upCount++;
								} else if (repositionValid(dummyNodes.get(i + 1), currentX, xcoords)) {
									upXPositions[i] = currentX;
									upCount++;
									// Leave currentX at same value
								} else {
									upXPositions[i] = nextX;
									currentX = nextX;
								}
							}

							currentX = xcoords.get(dummyNodes.get(numDummies - 1));

							for (int i = numDummies - 1; i > 0; i--) {
								float nextX = xcoords.get(dummyNodes.get(i - 1));

								if (currentX == nextX) {
									downXPositions[i - 1] = currentX;
									downCount++;
								} else if (repositionValid(dummyNodes.get(i - 1), currentX, xcoords)) {
									downXPositions[i - 1] = currentX;
									downCount++;
									// Leave currentX at same value
								} else {
									downXPositions[i - 1] = nextX;
									currentX = nextX;
								}
							}

							if (downCount <= count && upCount <= count) {
								// Neither of the new calculation provide a
								// straighter
								// edge
								continue;
							}

							if (downCount >= upCount) {
								// Apply down calculation values
								for (int i = numDummies - 2; i >= 0; i--) {
									// dummyNodes.get(i).putData(COORDINATE_X_HOLDER_KEY,
									// downXPositions[i]);
									xcoords.put(dummyNodes.get(i), downXPositions[i]);
								}
							} else if (upCount > downCount) {
								// Apply up calculation values
								for (int i = 1; i < numDummies; i++) {
									// dummyNodes.get(i).putData(COORDINATE_X_HOLDER_KEY,
									// upXPositions[i - 1]);
									xcoords.put(dummyNodes.get(i), upXPositions[i - 1]);
								}
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Determines whether or not a node may be moved to the specified x position
	 * on the specified rank
	 * 
	 * @return whether or not the virtual node can be moved to this position
	 */
	private boolean repositionValid(Node node, float position, Map<Node, Float> xcoords) {
		LayerContainer layerContainer = getContext().getLayerContainer();
		int layer = layerContainer.getLayer(node);
		List<Node> layerNodes = layerContainer.getNodes(layer);
		int layerIndex = layerContainer.getIndex(node);

		float currentX = xcoords.get(node);
		if (position < currentX) {
			if (layerIndex == 0) {
				return true;
			}

			Node leftNode = layerNodes.get(layerIndex - 1);
			float leftLimit = xcoords.get(leftNode);
			leftLimit = leftLimit + getWidth(leftNode) / 2 + getContext().getIntracellSpacing() + getWidth(node) / 2;

			if (leftLimit <= position) {
				return true;
			} else {
				return false;
			}
		} else if (position > currentX) {
			if (layerIndex == layerNodes.size() - 1) {
				return true;
			}

			Node rightNode = layerNodes.get(layerIndex + 1);
			float rightLimit = xcoords.get(rightNode);
			rightLimit = rightLimit - getWidth(rightNode) / 2 - getContext().getIntracellSpacing() - getWidth(node) / 2;

			if (rightLimit >= position) {
				return true;
			} else {
				return false;
			}
		}

		return true;
	}

	private void minNode(int i, Map<Node, Float> xcoords) {
		LayerContainer layerContainer = getContext().getLayerContainer();
		// Queue all nodes
		LinkedList<Node> queue = new LinkedList<Node>();

		boolean downward = (i % 2 == 0);
		if (downward) {
			for (int j = 0; j < layerContainer.getSize(); j++) {
				List<Node> layerNodes = layerContainer.getNodes(j);
				for (Node n : layerNodes) {
					queue.add(n);
					setVisitedNode(n, true);
				}
			}
		} else {
			for (int j = layerContainer.getSize() - 1; j >= 0; j--) {
				List<Node> layerNodes = layerContainer.getNodes(j);
				for (Node n : layerNodes) {
					queue.add(n);
					setVisitedNode(n, true);
				}
			}
		}

		// Set a limit of the maximum number of times we will access the queue
		// in case a loop appears
		int maxTries = queue.size() * 20;
		int count = 0;

		// Don't move cell within this value of their median
		int tolerance = 0;

		while (!queue.isEmpty() && count <= maxTries) {
			Node node = queue.getFirst();

			int layer = layerContainer.getLayer(node);
			int index = layerContainer.getIndex(node);

			List<Node> nextLayerNeighbors = layerContainer.getNeighbors(node, +1);
			List<Node> previousLayerNeighbors = layerContainer.getNeighbors(node, -1);

			int nextLayerNeighborsSize = nextLayerNeighbors.size();
			int previousLayerNeighborsSize = previousLayerNeighbors.size();
			int totalConnectedNeighbors = nextLayerNeighborsSize + previousLayerNeighborsSize;

			float nextLayerMedian = calculateMedianPosition(nextLayerNeighbors, xcoords);
			float previousLayerMedian = calculateMedianPosition(previousLayerNeighbors, xcoords);
			float currentPosition = xcoords.get(node);
			float nodeMedian = currentPosition;

			if (totalConnectedNeighbors > 0) {
				nodeMedian = (nextLayerMedian * nextLayerNeighborsSize
						+ previousLayerMedian * previousLayerNeighborsSize) / totalConnectedNeighbors;
			}

			// Flag storing whether or not position has changed
			boolean positionChanged = false;

			if (nodeMedian < currentPosition - tolerance) {
				if (index == 0) {
					xcoords.put(node, nodeMedian);
					positionChanged = true;
				} else {
					Node leftNode = layerContainer.getNodes(layer).get(index - 1);
					float leftLimit = xcoords.get(leftNode);
					leftLimit = leftLimit + getWidth(leftNode) / 2 + getContext().getIntracellSpacing()
							+ getWidth(node) / 2;

					if (leftLimit < nodeMedian) {
						xcoords.put(node, nodeMedian);
						positionChanged = true;
					} else if (leftLimit < xcoords.get(node) - tolerance) {
						xcoords.put(node, leftLimit);
						positionChanged = true;
					}
				}
			} else if (nodeMedian > currentPosition + tolerance) {
				int layerSize = layerContainer.getNodes(layer).size();

				if (index == layerSize - 1) {
					xcoords.put(node, nodeMedian);
					positionChanged = true;
				} else {
					Node rightNode = layerContainer.getNodes(layer).get(index + 1);
					float rightLimit = xcoords.get(rightNode);
					rightLimit = rightLimit - getWidth(rightNode) / 2 - getContext().getIntracellSpacing()
							- getWidth(node) / 2;

					if (rightLimit > nodeMedian) {
						xcoords.put(node, nodeMedian);
						positionChanged = true;
					} else if (rightLimit > xcoords.get(node) + tolerance) {
						xcoords.put(node, rightLimit);
						positionChanged = true;
					}
				}
			}

			if (positionChanged) {
				// Add connected nodes to map and list
				for (Node neighbor : nextLayerNeighbors) {
					if (!isVisited(neighbor)) {
						setVisitedNode(neighbor, true);
						if (!queue.contains(neighbor)) {
							queue.add(neighbor);
						}
					}
				}

				// Add connected nodes to map and list
				for (Node neighbor : previousLayerNeighbors) {
					if (!isVisited(neighbor)) {
						setVisitedNode(neighbor, true);
						if (!queue.contains(neighbor)) {
							queue.add(neighbor);
						}
					}
				}
			}

			queue.removeFirst();
			setVisitedNode(node, false);
			count++;
		}
	}

	private void minEdge(int i, Map<Node, Float> xcoords) {

	}

	private void medianPos(int i, Map<Node, Float> xcoords) {
		// Reverse sweep direction each time through this method
		boolean downwardSweep = (i % 2 == 0);

		if (downwardSweep) {
			for (int j = 1; j < getContext().getLayerContainer().getSize(); j++) {
				rankMedianPosition(j, -1, xcoords);
			}
		} else {
			for (int j = getContext().getLayerContainer().getSize() - 2; j >= 0; j--) {
				rankMedianPosition(j, +1, xcoords);
			}
		}
	}

	private void rankMedianPosition(int currentLayer, int nextLayer, Map<Node, Float> xcoords) {
		LayerContainer layerContainer = getContext().getLayerContainer();
		// Sort nodes by priority (priority is given by weighted sum of in (out)
		// edges) and initialize some variables
		List<Node> layerNodes = layerContainer.getNodes(currentLayer);
		List<Node> layerNodesCopy = new ArrayList<Node>(layerNodes);
		for (Node node : layerNodesCopy) {
			setVisitedNode(node, false);
		}
		comparator.setDownward(nextLayer < 0);
		Collections.sort(layerNodesCopy, comparator);

		for (int i = 0; i < layerNodesCopy.size(); i++) {
			Node node = layerNodesCopy.get(i);
			List<Node> neighbors = layerContainer.getNeighbors(node, nextLayer);
			float median;
			if (neighbors.size() > 0) {
				median = calculateMedianPosition(neighbors, xcoords);
			} else {
				median = xcoords.get(node);
			}

			// Try to place the node in its median position
			float leftBuffer = 0.0f;
			float leftLimit = -100000000.0f;

			for (int j = layerContainer.getIndex(node) - 1; j >= 0;) {
				Node leftNode = layerNodes.get(j);

				if (isVisited(leftNode)) {
					leftLimit = xcoords.get(leftNode) + getWidth(leftNode) / 2.0f + getContext().getIntracellSpacing()
							+ leftBuffer + getWidth(node) / 2.0f;
					j = -1;
				} else {
					leftBuffer += getWidth(leftNode) + getContext().getIntracellSpacing();
					j--;
				}
			}

			float rightBuffer = 0.0f;
			float rightLimit = 100000000.0f;

			for (int j = layerContainer.getIndex(node) + 1; j < layerNodes.size();) {
				Node rightNode = layerNodes.get(j);

				if (isVisited(rightNode)) {
					rightLimit = xcoords.get(rightNode) - getWidth(rightNode) / 2.0f
							- getContext().getIntracellSpacing() - rightBuffer - getWidth(rightNode) / 2.0f;
					j = layerNodes.size();
				} else {
					rightBuffer += getWidth(rightNode) + getContext().getIntracellSpacing();
					j++;
				}
			}

			if (median >= leftLimit && median <= rightLimit) {
				xcoords.put(node, median);
			} else if (median < leftLimit) {
				// Couldn't place at median value, place as close to that
				// value as possible
				xcoords.put(node, leftLimit);
			} else if (median > rightLimit) {
				// Couldn't place at median value, place as close to that
				// value as possible
				xcoords.put(node, rightLimit);
			}

			setVisitedNode(node, true);
		}
	}

	private float calculateMedianPosition(List<Node> neighbors, Map<Node, Float> xcoords) {
		float median = 0;

		float[] positions = new float[neighbors.size()];
		for (int i = 0; i < positions.length; i++) {
			positions[i] = xcoords.get(neighbors.get(i));
		}
		Arrays.sort(positions);

		if (positions.length > 0) {
			if (positions.length % 2 == 0) {
				median = (positions[positions.length / 2] + positions[positions.length / 2 - 1]) / 2;
			} else {
				median = positions[positions.length / 2];
			}
		}

		return median;
	}

	private Map<Node, Float> initXCoords(Graph graph) {
		Map<Node, Float> xcoords = new HashMap<>();

		for (int i = 0; i < getContext().getLayerContainer().getSize(); i++) {
			float position = 0f;
			List<Node> layerNodes = getContext().getLayerContainer().getNodes(i);
			Node u = layerNodes.get(0);
			xcoords.put(u, position);

			for (int j = 1; j < layerNodes.size(); j++) {
				Node v = layerNodes.get(j);
				position += separation(u, v);
				xcoords.put(v, position);
				u = v;
			}
		}

		return xcoords;
	}

	private float separation(Node u, Node v) {
		return ((xsize(u) + xsize(v)) / 2) + getContext().getIntracellSpacing();
	}

	private class WeightedComparator implements Comparator<Node> {

		private boolean downward;

		@Override
		public int compare(Node u, Node v) {
			if (downward) {
				return v.getInDegree() - u.getInDegree();
			} else {
				return v.getOutDegree() - u.getOutDegree();
			}
		}

		public void setDownward(boolean downward) {
			this.downward = downward;
		}
	}
}
