package org.processmining.humbertogarciacaballero.models.model.layout.stage.emplacer.port;

import java.util.Iterator;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.LayoutStage;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;

public class PortCalculator extends LayoutStage {

    public PortCalculator(HierarchicalLayoutContext context) {
        super(context);
    }

    @SuppressWarnings("unchecked")
    @Override
    public int run(Graph graph) {
        Iterator<Node> iter = graph.nodes();
        while (iter.hasNext()) {
            Node node = iter.next();
            // calculatePorts(graph, node);
            calculatePortsNewVersion(graph, node);
        }
        return 0;
    }

    public void calculatePortsNewVersion(Graph graph, Node node) {
        List<Node> nextNeighbors = getContext().getLayerContainer().getNeighbors(node, +1);
        List<Node> sameLayerNeighbors = getContext().getLayerContainer().getNeighbors(node, 0);

        int neighborsNum = nextNeighbors.size() + sameLayerNeighbors.size()
                + calculateBidirectionalEdges(graph, node, nextNeighbors)
                + calculateBidirectionalEdges(graph, node, sameLayerNeighbors);
        float size = xsize(node);

        if (neighborsNum > 0) {
            float distance = Math.min((size / neighborsNum), getContext().getPortSeparation());
            float currentPosition = (((neighborsNum / 2)) * distance) + (distance / (2 - (neighborsNum % 2)));
            int nodeIndex = getRelativeIndex(node, sameLayerNeighbors);

            // Nodes in the same layer and below this node
            for (int i = nodeIndex - 1; i >= 0; i--) {
                Node neighbor = sameLayerNeighbors.get(i);
                Edge e = graph.getEdge(node, neighbor);
                float offset = currentPosition -= distance;
                if (e == null) {
                    e = graph.getEdge(neighbor, node);
                    setTargetOffset(e, offset);
                    if (getEdgeParams(e).bidirectionalCycles) {
                        offset = currentPosition /*-= distance*/;
                        setBidirectionalTargetOffset(e, offset);
                    }
                } else {
                    setSourceOffset(e, offset);
                    if (getEdgeParams(e).bidirectionalCycles) {
                        offset = currentPosition /*-= distance*/;
                        setBidirectionalSourceOffset(e, offset);
                    }
                }
            }

            // Nodes in the next layer
            for (int i = 0; i < nextNeighbors.size(); i++) {
                Node neighbor = nextNeighbors.get(i);
                Edge e = graph.getEdge(node, neighbor);
                float offset = currentPosition -= distance;
                if (e == null) {
                    e = graph.getEdge(neighbor, node);
                    setTargetOffset(e, offset);
                    if (getEdgeParams(e).bidirectionalCycles) {
                        offset = currentPosition /*-= distance*/;
                        setBidirectionalTargetOffset(e, offset);
                    }
                } else {
                    setSourceOffset(e, offset);
                    if (getEdgeParams(e).bidirectionalCycles) {
                        offset = currentPosition /*-= distance*/;
                        setBidirectionalSourceOffset(e, offset);
                    }
                }
            }

            // Nodes in the same layer and above this node
            for (int i = sameLayerNeighbors.size() - 1; i >= nodeIndex; i--) {
                Node neighbor = sameLayerNeighbors.get(i);
                Edge e = graph.getEdge(node, neighbor);
                float offset = currentPosition -= distance;
                if (e == null) {
                    e = graph.getEdge(neighbor, node);
                    setTargetOffset(e, offset);
                    if (getEdgeParams(e).bidirectionalCycles) {
                        offset = currentPosition /*-= distance*/;
                        setBidirectionalTargetOffset(e, offset);
                    }
                } else {
                    setSourceOffset(e, offset);
                    if (getEdgeParams(e).bidirectionalCycles) {
                        offset = currentPosition /*-= distance*/;
                        setBidirectionalSourceOffset(e, offset);
                    }
                }
            }
        }

        // Previous layer nodes
        nextNeighbors = getContext().getLayerContainer().getNeighbors(node, -1);
        neighborsNum = nextNeighbors.size() + calculateBidirectionalEdges(graph, node, nextNeighbors);
        if (neighborsNum > 0) {
            float distance = Math.min((size / neighborsNum), getContext().getPortSeparation());
            float currentPosition = (((neighborsNum / 2)) * distance) + (distance / (2 - (neighborsNum % 2)));

            for (int i = 0; i < nextNeighbors.size(); i++) {
                Node neighbor = nextNeighbors.get(i);
                Edge e = graph.getEdge(node, neighbor);
                if (e == null) {
                    e = graph.getEdge(neighbor, node);
                    setTargetOffset(e, currentPosition -= distance);
                    if (getEdgeParams(e).bidirectionalCycles) {
                        setBidirectionalTargetOffset(e, currentPosition /*-= distance*/);
                    }
                } else {
                    setSourceOffset(e, currentPosition -= distance);
                    if (getEdgeParams(e).bidirectionalCycles) {
                        setBidirectionalSourceOffset(e, currentPosition /*-= distance*/);
                    }
                }
            }
        }
    }

//    public void calculatePorts(Graph graph, Node node) {
//        List<Node> nextNeighbors = getContext().getLayerContainer().getNeighbors(node, +1);
//        List<Node> sameLayerNeighbors = getContext().getLayerContainer().getNeighbors(node, 0);
//
//        if (NodeType.PLACE.equals(NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY)))
//                || !getContext().isSimplifierActive()) {
//            int neighborsNum = nextNeighbors.size() + sameLayerNeighbors.size()
//                    + calculateBidirectionalEdges(graph, node, nextNeighbors)
//                    + calculateBidirectionalEdges(graph, node, sameLayerNeighbors);
//            float size = xsize(node);
//
//            if (neighborsNum > 0) {
//                float distance = Math.min((size / neighborsNum), getContext().getPortSeparation());
//                float currentPosition = (((neighborsNum / 2)) * distance) + (distance / (2 - (neighborsNum % 2)));
//                int nodeIndex = getRelativeIndex(node, sameLayerNeighbors);
//
//                // Nodes in the same layer and below this node
//                for (int i = nodeIndex - 1; i >= 0; i--) {
//                    Node neighbor = sameLayerNeighbors.get(i);
//                    Edge e = graph.getEdge(node, neighbor);
//                    float offset = currentPosition -= distance;
//                    if (e == null) {
//                        e = graph.getEdge(neighbor, node);
//                        setTargetOffset(e, offset);
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            offset = currentPosition -= distance;
//                            setBidirectionalTargetOffset(e, offset);
//                        }
//                    } else {
//                        setSourceOffset(e, offset);
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            offset = currentPosition -= distance;
//                            setBidirectionalSourceOffset(e, offset);
//                        }
//                    }
//                }
//
//                // Nodes in the next layer
//                for (int i = 0; i < nextNeighbors.size(); i++) {
//                    Node neighbor = nextNeighbors.get(i);
//                    Edge e = graph.getEdge(node, neighbor);
//                    float offset = currentPosition -= distance;
//                    if (e == null) {
//                        e = graph.getEdge(neighbor, node);
//                        setTargetOffset(e, offset);
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            offset = currentPosition -= distance;
//                            setBidirectionalTargetOffset(e, offset);
//                        }
//                    } else {
//                        setSourceOffset(e, offset);
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            offset = currentPosition -= distance;
//                            setBidirectionalSourceOffset(e, offset);
//                        }
//                    }
//                }
//
//                // Nodes in the same layer and above this node
//                for (int i = sameLayerNeighbors.size() - 1; i >= nodeIndex; i--) {
//                    Node neighbor = sameLayerNeighbors.get(i);
//                    Edge e = graph.getEdge(node, neighbor);
//                    float offset = currentPosition -= distance;
//                    if (e == null) {
//                        e = graph.getEdge(neighbor, node);
//                        setTargetOffset(e, offset);
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            offset = currentPosition -= distance;
//                            setBidirectionalTargetOffset(e, offset);
//                        }
//                    } else {
//                        setSourceOffset(e, offset);
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            offset = currentPosition -= distance;
//                            setBidirectionalSourceOffset(e, offset);
//                        }
//                    }
//                }
//            }
//
//            // Previous layer nodes
//            nextNeighbors = getContext().getLayerContainer().getNeighbors(node, -1);
//            neighborsNum = nextNeighbors.size() + calculateBidirectionalEdges(graph, node, nextNeighbors);
//            if (neighborsNum > 0) {
//                float distance = Math.min((size / neighborsNum), getContext().getPortSeparation());
//                float currentPosition = (((neighborsNum / 2)) * distance) + (distance / (2 - (neighborsNum % 2)));
//
//                for (int i = 0; i < nextNeighbors.size(); i++) {
//                    Node neighbor = nextNeighbors.get(i);
//                    Edge e = graph.getEdge(node, neighbor);
//                    if (e == null) {
//                        e = graph.getEdge(neighbor, node);
//                        setTargetOffset(e, currentPosition -= distance);
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            setBidirectionalTargetOffset(e, currentPosition -= distance);
//                        }
//                    } else {
//                        setSourceOffset(e, currentPosition -= distance);
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            setBidirectionalSourceOffset(e, currentPosition -= distance);
//                        }
//                    }
//                }
//            }
//        } else {
//            int neighborsNum = countSimplifiedNeighbors(node, nextNeighbors)
//                    + countSimplifiedNeighbors(node, sameLayerNeighbors);
//            float size = xsize(node);
//
//            if (neighborsNum > 0) {
//                List<Node> allNeighbors = new ArrayList<Node>(nextNeighbors);
//                allNeighbors.addAll(sameLayerNeighbors);
//                float distance = Math.min((size / neighborsNum), getContext().getPortSeparation());
//                float currentPosition = (((neighborsNum / 2)) * distance) + (distance / (2 - (neighborsNum % 2)));
//                float simplifiedPosition = currentPosition - ((neighborsNum / 2) * distance + distance);
//                int nodeIndex = getRelativeIndex(node, sameLayerNeighbors);
//                int nonSimplifiedNodes = 0;
//                int simplifiedIndex = isSimplifiedNeighbors(node, allNeighbors) ? (int) Math.ceil(neighborsNum / 2)
//                        : -1;
//
//                // Nodes in the same layer and below this node
//                for (int i = nodeIndex - 1; i >= 0; i--) {
//                    Node neighbor = sameLayerNeighbors.get(i);
//                    Edge e = graph.getEdge(node, neighbor);
//                    float offset;
//
//                    if (e == null) {
//                        e = graph.getEdge(neighbor, node);
//
//                        if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                            offset = simplifiedPosition;
//                        } else {
//                            offset = currentPosition -= distance;
//                            if (nonSimplifiedNodes == simplifiedIndex)
//                                offset = currentPosition -= distance;
//                            nonSimplifiedNodes++;
//                        }
//
//                        setTargetOffset(e, offset);
//
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                                offset = simplifiedPosition;
//                            } else {
//                                offset = currentPosition -= distance;
//                                if (nonSimplifiedNodes == simplifiedIndex)
//                                    offset = currentPosition -= distance;
//                                nonSimplifiedNodes++;
//                            }
//                            setBidirectionalTargetOffset(e, offset);
//                        }
//                    } else {
//                        if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                            offset = simplifiedPosition;
//                        } else {
//                            offset = currentPosition -= distance;
//                            if (nonSimplifiedNodes == simplifiedIndex)
//                                offset = currentPosition -= distance;
//                            nonSimplifiedNodes++;
//                        }
//
//                        setSourceOffset(e, offset);
//
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                                offset = simplifiedPosition;
//                            } else {
//                                offset = currentPosition -= distance;
//                                if (nonSimplifiedNodes == simplifiedIndex)
//                                    offset = currentPosition -= distance;
//                                nonSimplifiedNodes++;
//                            }
//                            setBidirectionalSourceOffset(e, offset);
//                        }
//                    }
//                }
//
//                // Nodes in the next layer
//                for (int i = 0; i < nextNeighbors.size(); i++) {
//                    Node neighbor = nextNeighbors.get(i);
//                    Edge e = graph.getEdge(node, neighbor);
//                    float offset;
//                    if (e == null) {
//                        e = graph.getEdge(neighbor, node);
//
//                        if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                            offset = simplifiedPosition;
//                        } else {
//                            offset = currentPosition -= distance;
//                            if (nonSimplifiedNodes == simplifiedIndex)
//                                offset = currentPosition -= distance;
//                            nonSimplifiedNodes++;
//                        }
//
//                        setTargetOffset(e, offset);
//
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                                offset = simplifiedPosition;
//                            } else {
//                                offset = currentPosition -= distance;
//                                if (nonSimplifiedNodes == simplifiedIndex)
//                                    offset = currentPosition -= distance;
//                                nonSimplifiedNodes++;
//                            }
//                            setBidirectionalTargetOffset(e, offset);
//                        }
//                    } else {
//                        if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                            offset = simplifiedPosition;
//                        } else {
//                            offset = currentPosition -= distance;
//                            if (nonSimplifiedNodes == simplifiedIndex)
//                                offset = currentPosition -= distance;
//                            nonSimplifiedNodes++;
//                        }
//
//                        setSourceOffset(e, offset);
//
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                                offset = simplifiedPosition;
//                            } else {
//                                offset = currentPosition -= distance;
//                                if (nonSimplifiedNodes == simplifiedIndex)
//                                    offset = currentPosition -= distance;
//                                nonSimplifiedNodes++;
//                            }
//                            setBidirectionalSourceOffset(e, offset);
//                        }
//                    }
//                }
//
//                // Nodes in the same layer and above this node
//                for (int i = sameLayerNeighbors.size() - 1; i >= nodeIndex; i--) {
//                    Node neighbor = sameLayerNeighbors.get(i);
//                    Edge e = graph.getEdge(node, neighbor);
//                    float offset;
//                    if (e == null) {
//                        e = graph.getEdge(neighbor, node);
//                        if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                            offset = simplifiedPosition;
//                        } else {
//                            offset = currentPosition -= distance;
//                            if (nonSimplifiedNodes == simplifiedIndex)
//                                offset = currentPosition -= distance;
//                            nonSimplifiedNodes++;
//                        }
//                        setTargetOffset(e, offset);
//
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                                offset = simplifiedPosition;
//                            } else {
//                                offset = currentPosition -= distance;
//                                if (nonSimplifiedNodes == simplifiedIndex)
//                                    offset = currentPosition -= distance;
//                                nonSimplifiedNodes++;
//                            }
//                            setBidirectionalTargetOffset(e, offset);
//                        }
//                    } else {
//                        if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                            offset = simplifiedPosition;
//                        } else {
//                            offset = currentPosition -= distance;
//                            if (nonSimplifiedNodes == simplifiedIndex)
//                                offset = currentPosition -= distance;
//                            nonSimplifiedNodes++;
//                        }
//                        setSourceOffset(e, offset);
//
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                                offset = simplifiedPosition;
//                            } else {
//                                offset = currentPosition -= distance;
//                                if (nonSimplifiedNodes == simplifiedIndex)
//                                    offset = currentPosition -= distance;
//                                nonSimplifiedNodes++;
//                            }
//                            setBidirectionalSourceOffset(e, offset);
//                        }
//                    }
//                }
//            }
//
//            // Previous layer nodes
//            nextNeighbors = getContext().getLayerContainer().getNeighbors(node, -1);
//            neighborsNum = countSimplifiedNeighbors(node, nextNeighbors);
//            if (neighborsNum > 0) {
//                float distance = Math.min((size / neighborsNum), getContext().getPortSeparation());
//                float currentPosition = (((neighborsNum / 2)) * distance) + (distance / (2 - (neighborsNum % 2)));
//                float simplifiedPosition = currentPosition - ((neighborsNum / 2) * distance + distance);
//                int nonSimplifiedNodes = 0;
//                int simplifiedIndex = isSimplifiedNeighbors(node, nextNeighbors) ? (int) Math.ceil(neighborsNum / 2)
//                        : -1;
//
//                for (int i = 0; i < nextNeighbors.size(); i++) {
//                    Node neighbor = nextNeighbors.get(i);
//                    Edge e = graph.getEdge(node, neighbor);
//                    float offset;
//                    if (e == null) {
//                        e = graph.getEdge(neighbor, node);
//
//                        if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                            offset = simplifiedPosition;
//                        } else {
//                            offset = currentPosition -= distance;
//                            if (nonSimplifiedNodes == simplifiedIndex)
//                                offset = currentPosition -= distance;
//                            nonSimplifiedNodes++;
//                        }
//
//                        setTargetOffset(e, offset);
//
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                                offset = simplifiedPosition;
//                            } else {
//                                offset = currentPosition -= distance;
//                                if (nonSimplifiedNodes == simplifiedIndex)
//                                    offset = currentPosition -= distance;
//                                nonSimplifiedNodes++;
//                            }
//                            setBidirectionalTargetOffset(e, offset);
//                        }
//                    } else {
//
//                        if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                            offset = simplifiedPosition;
//                        } else {
//                            offset = currentPosition -= distance;
//                            if (nonSimplifiedNodes == simplifiedIndex)
//                                offset = currentPosition -= distance;
//                            nonSimplifiedNodes++;
//                        }
//
//                        setSourceOffset(e, offset);
//
//                        if (getEdgeParams(e).bidirectionalCycles) {
//                            if (HierarchicalLayouter.getEdgeParams(e).simplified) {
//                                offset = simplifiedPosition;
//                            } else {
//                                offset = currentPosition -= distance;
//                                if (nonSimplifiedNodes == simplifiedIndex)
//                                    offset = currentPosition -= distance;
//                                nonSimplifiedNodes++;
//                            }
//                            setBidirectionalSourceOffset(e, offset);
//                        }
//                    }
//                }
//            }
//        }
//    }

    private int calculateBidirectionalEdges(Graph graph, Node node, List<Node> neighbors) {
        int bidirectionalEdges = 0;

//        for (Node neighbor : neighbors) {
//            Edge e = graph.getEdge(node, neighbor);
//            if (e == null) {
//                e = graph.getEdge(neighbor, node);
//            }
//            if (getEdgeParams(e).bidirectionalCycles)
//                bidirectionalEdges++;
//        }

        return bidirectionalEdges;
    }

    private boolean isSimplifiedNeighbors(Node source, List<Node> neighbors) {
        boolean simplifiedFlag = false;
        Graph graph = source.getGraph();
        for (int i = 0; i < neighbors.size() && !simplifiedFlag; i++) {
            Node neighbor = neighbors.get(i);
            Edge edge = graph.getEdge(source, neighbor);
            if (edge == null) {
                edge = graph.getEdge(neighbor, source);
            }

            simplifiedFlag = HierarchicalLayouter.getEdgeParams(edge).simplified;
        }

        return simplifiedFlag;
    }

    /**
     * Counts the number of neighbors of source which have not been simplified.
     * In the case that there is at least one simplified edge, the count is
     * increased by 1.
     * 
     * @param source
     *            Source node.
     * @param neighbors
     *            Neighbors of this node.
     * @return Count of non simplified neighbors and (optionally) plus 1.
     */
    private int countSimplifiedNeighbors(Node source, List<Node> neighbors) {
        int count = 0;
        boolean simplifiedFlag = false;
        Graph graph = source.getGraph();
        for (Node neighbor : neighbors) {
            Edge edge = graph.getEdge(source, neighbor);
            if (edge == null) {
                edge = graph.getEdge(neighbor, source);
            }

            if (HierarchicalLayouter.getEdgeParams(edge).simplified) {
                simplifiedFlag = true;
            } else {
                count++;
                if (getEdgeParams(edge).bidirectionalCycles)
                    count++;
            }
        }

        if (simplifiedFlag) {
            count++;
        }

        return count;
    }

    private int getRelativeIndex(Node node, List<Node> neighbors) {
        int nodeIndex = getContext().getLayerContainer().getIndex(node);
        int index = 0;
        while (index < neighbors.size()
                && nodeIndex > getContext().getLayerContainer().getIndex(neighbors.get(index))) {
            index++;
        }

        return index;
    }

    private void setSourceOffset(Edge e, float offset) {
        HierarchicalLayouter.getEdgeParams(e).sourceOffset = offset;
    }

    private void setTargetOffset(Edge e, float offset) {
        HierarchicalLayouter.getEdgeParams(e).targetOffset = offset;
    }

    private void setBidirectionalSourceOffset(Edge e, float offset) {
        HierarchicalLayouter.getEdgeParams(e).bidirectionalSourceOffset = offset;
    }

    private void setBidirectionalTargetOffset(Edge e, float offset) {
        HierarchicalLayouter.getEdgeParams(e).bidirectionalTargetOffset = offset;
    }

    protected float xsize(Node u) {
        switch (getContext().getLayoutOrientation()) {
        case TOP_TO_BOTTOM:
        case BOTTOM_TO_TOP:
            return getWidth(u);
        default:
            return getHeight(u);
        }
    }

    protected float getWidth(Node n) {
        return HierarchicalLayouter.getNodeParams(n).width;
    }

    protected float getHeight(Node n) {
        return HierarchicalLayouter.getNodeParams(n).height;
    }
}
