package org.processmining.humbertogarciacaballero.models.model.layout.stage.layerer;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;

import com.google.common.collect.Sets;
import com.google.common.collect.Sets.SetView;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;

// https://cs.brown.edu/~rt/gdhandbook/chapters/hierarchical.pdf
public class LongestPathLayerer extends Layerer {

	private static final int SOURCESCANSTARTRANK = 100000000;
	private int maxRank;

	public LongestPathLayerer(HierarchicalLayoutContext context) {
		super(context);
	}

	@SuppressWarnings("unchecked")
	@Override
	public void calculateLayers(Graph graph) {
		Set<Node> V = Sets.newHashSet(graph.nodes());
		Set<Node> U = new HashSet<>();
		Set<Node> Z = new HashSet<>();
		int currentLayer = 0;

		while (U.size() != graph.getNodeCount()) {
			Node v = selectVertex(V, U, Z);
			if (v != null) {
				setLayer(v, currentLayer);
				U.add(v);
			} else {
				currentLayer = currentLayer + 1;
				Z.addAll(U);
			}
		}

		//        vertexPromotion(graph);

		Iterator<Edge> edges = graph.edges();
		while (edges.hasNext()) {
			Edge e = edges.next();
			int distance = getContext().getLayerContainer().getDistance(e);
			if (distance > 1) {
				createDummies(graph, e.getSourceNode(), e.getTargetNode());
			}
		}
//		maxRank = 0;
//		Set<Edge> edges = new HashSet<>();
//		Set<Node> nnodes = new HashSet<>();
//		LinkedList<Node> startNodes = Lists.newLinkedList(Lists.newArrayList(getSinkNodes(graph)));
//		HashMap<Node, Integer> layers = new HashMap<>();
//		Iterator<Node> iter = graph.nodes();
//		while (iter.hasNext()) {
//			Node n = iter.next();
//			layers.put(n, -1);
//		}
//
//		List<Node> startNodesCopy = new ArrayList<Node>(startNodes);
//		while (!startNodes.isEmpty()) {
//			Node internalNode = startNodes.get(0);
//			Iterator<Edge> iter2;
//			Iterator<Edge> iter3;
//			iter2 = internalNode.outEdges();
//			iter3 = internalNode.inEdges();
//
//			// flag to keep track of whether or not all layer determining
//			// edges have been scanned
//			boolean allEdgesScanned = true;
//			// Work out the layer of this node from the layer determining
//			// edges
//			// The minimum layer number of any node connected by one of
//			// the layer determining edges variable. If we are starting
//			// from sources, need to start at some huge value and
//			// normalise down afterwards
//			int minimumLayer = 0;
//			minimumLayer = SOURCESCANSTARTRANK;
//			while (allEdgesScanned && iter2.hasNext()) {
//				Edge internalEdge = iter2.next();
//				if (edges.contains(internalEdge)) {
//					// This edge has been scanned, get the layer of the
//					// node on the other end
//					Node otherNode = internalEdge.getTargetNode();
//					minimumLayer = Math.max(minimumLayer, layers.get(otherNode) + 1);
//				} else {
//					allEdgesScanned = false;
//				}
//			}
//			// If all edge have been scanned, assign the layer, mark all
//			// edges in the other direction and remove from the nodes list
//			if (allEdgesScanned) {
//				layers.put(internalNode, minimumLayer);
//				maxRank = Math.max(maxRank, minimumLayer);
//				while (iter3.hasNext()) {
//					Edge internalEdge = iter3.next();
//					// Assign unique stamp ( y/m/d/h )
//					edges.add(internalEdge);
//					// Add node on other end of edge to LinkedList of
//					// nodes
//					// to be analysed
//					Node otherNode = internalEdge.getSourceNode();
//					// Only add node if it hasn't been assigned a layer
//					if (layers.get(otherNode) == -1) {
//						startNodes.addLast(otherNode);
//						// Mark this other node as neither being
//						// unassigned nor assigned so it isn't
//						// added to this list again, but it's
//						// layer isn't used in any calculation.
//						nnodes.add(otherNode);
//					}
//				}
//				startNodes.removeFirst();
//			} else {
//				// Not all the edges have been scanned, get to the back of
//				// the class and put the dunces cap on
//				Object removedCell = startNodes.removeFirst();
//				startNodes.addLast(internalNode);
//				if (removedCell == internalNode && startNodes.size() == 1) {
//					// This is an error condition, we can't get out of
//					// this loop. It could happen for more than one node
//					// but that's a lot harder to detect. Log the error
//					// TODO make log comment
//					break;
//				}
//			}
//		}
//
//		// Tighten the rank 0 nodes as far as possible
//		for (Node internalNode : startNodesCopy) {
//			int currentMinLayer = 1000000;
//			Iterator<Node> iter2 = internalNode.inNeighbors();
//			while (iter2.hasNext()) {
//				Node otherNode = iter2.next();
//				layers.put(internalNode, Math.min(currentMinLayer, layers.get(otherNode) - 1));
//				currentMinLayer = layers.get(internalNode);
//			}
//		}
//		
//		iter = graph.nodes();
//		while (iter.hasNext()) {
//			Node n = iter.next();
//			setLayer(n, layers.get(n));
//		}
	}

	private Node selectVertex(Set<Node> V, Set<Node> U, Set<Node> Z) {
		SetView<Node> view = Sets.difference(V, U);
		for (Node node : view) {
			boolean allAssigned = true;
			Iterator<Edge> edges = node.inEdges();
			while (edges.hasNext()) {
				Edge e = edges.next();
				if (!Z.contains(e.getAdjacentNode(node))) {
					allAssigned = false;
					break;
				}
			}

			if (allAssigned)
				return node;
		}
		return null;
	}

	//    private int promoteVertex(Node v) {
	//        int dummydiff = 0;
	//        Iterator<Node> iter = v.inNeighbors();
	//        while (iter.hasNext()) {
	//            Node u = iter.next();
	//            if (getLayer(u) == getLayer(v) + 1) {
	//                dummydiff = dummydiff + promoteVertex(u);
	//            }
	//        }
	//        moveNode(v, getLayer(v) + 1);
	//        dummydiff = dummydiff - immediateNeighbors(getLayer(v), v.inNeighbors()) + immediateNeighbors(getLayer(v), v.outNeighbors());
	//        return dummydiff;
	//    }
	//    
	//    private int immediateNeighbors(int layer, Iterator<Node> iterator) {
	//        int sum = 0;
	//        while (iterator.hasNext()) {
	//            if (Math.abs(getLayer(iterator.next()) - layer) == 1) {
	//                sum++;
	//            }
	//        }
	//        return sum;
	//    }
	//    
	//    private void vertexPromotion(Graph graph) {
	//        int promotions;
	//        LayerContainer backup = new LayerContainer(getContext().getLayerContainer());
	//        do {
	//            promotions = 0;
	//            Iterator<Node> iter = graph.nodes();
	//            while (iter.hasNext()) {
	//                Node v = iter.next();
	//                if (v.getInDegree() > 0) {
	//                    if (promoteVertex(v) < 0) {
	//                        promotions = promotions + 1;
	//                        backup = new LayerContainer(getContext().getLayerContainer());
	//                    } else {
	//                        getContext().getLayerContainer().reset(backup);
	//                    }
	//                }
	//            }
	//        } while(promotions != 0);
	//    }    

	//    private void dfs(Graph graph, Node v, int layer) {
	//        setLayer(v, layer);
	//        Iterator<Node> iter = getRealPredecessors(v);
	//
	//        while (iter.hasNext()) {
	//            Node w = iter.next();
	//            if (getLayer(w) == -1) {
	//                dfs(graph, w, layer + 1);
	//            } else {
	//                createDummies(graph, v, w);
	//            }
	//        }
	//    }
}
