package org.processmining.humbertogarciacaballero.models.model.layout.stage.simplifier;

import java.util.Iterator;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.EdgeParams;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayerContainer;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.LayoutStage;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.util.ColorLib;
import prefuse.visual.VisualItem;

public class ColorMapper extends LayoutStage {

    private int[] palette;
    private int oldDifferentColors;

    public ColorMapper(HierarchicalLayoutContext context) {
        super(context);
        oldDifferentColors = 0;
    }

    public int run(Graph graph) {
        // Calculates how many different colors we need.
        LayerContainer layerContainer = getContext().getLayerContainer();
        int differentColors = 0;
        {
            Iterator<Node> iter = graph.nodes();
            while (iter.hasNext()) {
                Node node = iter.next();
                if (isColorable(node))
                    differentColors++;
            }
        }
        // for (int i = 0; i < layerContainer.getSize() - 1; i++) {
        // int betweenLayerColors = 0;
        // List<Node> layerNodes = new LinkedList<>(layerContainer.getNodes(i));
        // layerNodes.addAll(layerContainer.getNodes(i + 1));
        //
        // for (Node n : layerNodes) {
        // if (isColorable(n)) {
        // betweenLayerColors++;
        // }
        // }
        // differentColors = Math.max(differentColors, betweenLayerColors);
        // }

        if (oldDifferentColors != differentColors) {
            resetColorMap();
        }

        palette = ColorLib.getCategoryPalette(differentColors, 0.9f, 0.2f, 1.f, 1.0f);
        calculateColors(graph);

        // Assigns colors
        @SuppressWarnings("unchecked")
        Iterator<Edge> iter = graph.edges();
        while (iter.hasNext()) {
            Edge edge = iter.next();
            EdgeParams params = (EdgeParams) edge.get(HierarchicalLayouter.EDGE_PARAMS);
            if (params.simplified) {
                Node node;
                if (NodeType
                        .valueOf(edge.getSourceNode().getString(PNConstants.NODE_TYPE_KEY)) == NodeType.TRANSITION) {
                    node = edge.getSourceNode();
                } else {
                    node = edge.getTargetNode();
                }
                params.color = getColor(graph, node);
            } else if (edge.getBoolean(VisualItem.HIGHLIGHT)) {
                params.color = PNConstants.EDGE_HIGHLIGHT_COLOR;
            } else {
                params.color = PNConstants.EDGE_DEFAULT_COLOR;
            }
        }

        oldDifferentColors = differentColors;

        return 0;
    }

    private boolean isColorable(Node n) {
        return HierarchicalLayouter.getNodeParams(n).simplified
                && NodeType.valueOf(n.getString(PNConstants.NODE_TYPE_KEY)) == NodeType.TRANSITION;
    }

    private void calculateColors(Graph graph) {
        LayerContainer layerContainer = getContext().getLayerContainer();
        int index = 0;
        for (int i = 0; i < layerContainer.getSize(); i++) {
            List<Node> layerNodes = layerContainer.getNodes(i);
            for (Node node : layerNodes) {
                if (isColorable(node) && !getContext().getColorMap().containsKey(graph.getKey(node.getRow()))) {
                    int color = ColorLib.darker(palette[index % palette.length]);
                    getContext().getColorMap().put(graph.getKey(node.getRow()), color);
                    index++;
                }
            }
        }
    }

    private int getColor(Graph graph, Node node) {
        long key = graph.getKey(node.getRow());
        return getContext().getColorMap().get(key);
        // System.out.println("asdfasdf");
        //
        // // Looks for the appropriate color
        // int color = 0;
        // int layer = getContext().getLayerContainer().getLayer(node);
        // int index = 0;
        // List<Node> layerNodes;
        // if (layer == 0) {
        // layerNodes = getContext().getLayerContainer().getNodes(layer);
        // } else if (layer == 1) {
        // layerNodes = new
        // LinkedList<>(getContext().getLayerContainer().getNodes(layer - 1));
        // layerNodes.addAll(getContext().getLayerContainer().getNodes(layer));
        // } else {
        // layerNodes = new
        // LinkedList<>(getContext().getLayerContainer().getNodes(layer - 2));
        // layerNodes.addAll(getContext().getLayerContainer().getNodes(layer -
        // 1));
        // layerNodes.addAll(getContext().getLayerContainer().getNodes(layer));
        // }
        //
        // for (int i = 0; i < layerNodes.size() &&
        // !layerNodes.get(i).equals(node); i++) {
        // if (isColorable(layerNodes.get(i))) {
        // index++;
        // }
        // }
        //
        // color = ColorLib.darker(palette[index % palette.length]);
        // getContext().getColorMap().put(key, color);
        // return color;
    }

    public void resetColorMap() {
        getContext().getColorMap().clear();
    }
}
