package org.processmining.petrinettobpmn;

import java.awt.Color;
import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.JPanel;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.graphvisualizers.parameters.GraphVisualizerParameters;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.AttributeMap.ArrowType;
import org.processmining.models.graphbased.AttributeMapOwner;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.jgraph.ProMJGraph;
import org.processmining.models.jgraph.visualization.ProMJGraphPanel;
import org.processmining.plugins.graphviz.colourMaps.ColourMap;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;
import org.processmining.plugins.graphviz.dot.DotEdge;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.petrinet.PetriNetVisualization;

@Plugin(name = "Petrinet to BPMN convertor", parameterLabels = { "Petrinet ", }, returnLabels = {
		"Interactive Petri net" }, returnTypes = { JPanel.class })
@Visualizer
public class PetriNetToBPMNPlugin {

	/**
	 * The method that does the heavy lifting for your plug-in.
	 * 
	 * Note that this method only uses the boolean which is stored in the
	 * configuration. Nevertheless, it could have used the integer and/or the
	 * String as well.
	 * 
	 * @param context
	 *            The context where to run in.
	 * @param input1
	 *            The first input.
	 * @param dclist
	 * @param log
	 * @param input2
	 *            The second input.
	 * @param configuration
	 *            The configuration to use.
	 * @return The output.
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	static Set<String> events = new HashSet<String>();
	public static boolean stop = false;
	public static String GVPLACELABEL = "[GV]PlaceLabel";
	File dotFolder;
	PetrinetToBPMN pnbpmnmap;

	public Petrinet yourPrivatePlugin(UIPluginContext context, Petrinet pn) {
		int i = 0;
		Map<Transition, Integer> mapid = new HashMap<Transition, Integer>();
		for (Transition t : pn.getTransitions()) {
			if (t.isInvisible()) {
				mapid.put(t, i++);
			}
		}
		pnbpmnmap = new PetrinetToBPMN(pn);

		System.out.println(" And Splits : " + pnbpmnmap.getAndSplits().size() + " Xor : "
				+ pnbpmnmap.getXorSplits().size() + " IOR : " + pnbpmnmap.getIorSplits().size() + " Hidden : "
				+ pnbpmnmap.getHiddenNodes().size());

		return pn;
	}

	/**
	 * The plug-in variant that runs in a UI context and uses a dialog to get
	 * the configuration.
	 * 
	 * @param context
	 *            The context to run in.
	 * @param input1
	 *            The first input.
	 * @param input2
	 *            The second input.
	 * @return The output.
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	@UITopiaVariant(affiliation = "TU/e", author = "P.M. Dixit", email = "p.m.dixit@tue.nl")
	@PluginVariant(variantLabel = "PN to BPMN", requiredParameterLabels = { 0 })
	public JPanel yourDefaultPlugin(UIPluginContext context, Petrinet pn)
			throws IllegalAccessException, InstantiationException {
		yourPrivatePlugin(context, pn);

		GraphVisualizerParameters parameters = new GraphVisualizerParameters();
		PetriNetVisualization visualizer = new PetriNetVisualization();
		ProMJGraphPanel panel = (ProMJGraphPanel) visualizer.visualize(context, pn);
		ProMJGraph jGraph = panel.getGraph();
		ViewSpecificAttributeMap map = jGraph.getViewSpecificAttributes();
		for (Place place : pn.getPlaces()) {
			map.putViewSpecific(place, GVPLACELABEL, place.getLabel());
		}
		/*
		 * Got it. Now create the dot panel.
		 */
		Dot dot = apply(context, pn, map, parameters);
		DotPanel panel2 = new DotPanel(dot);

		return panel2;
	}

	public Dot apply(PluginContext context,
			DirectedGraph<? extends DirectedGraphNode, ? extends DirectedGraphEdge<?, ?>> graph,
			ViewSpecificAttributeMap map, GraphVisualizerParameters parameters) {
		Dot dot = new Dot() {
			public String toString() {
				return "strict " + super.toString();
			}
		};
		Map<DirectedGraphNode, DotNode> nodeMap = new HashMap<DirectedGraphNode, DotNode>();
		for (DirectedGraphNode node : graph.getNodes()) {
			if (!pnbpmnmap.hiddenNodes.contains(node)) {
				DotNode dotNode = dot.addNode(node.getLabel());
				nodeMap.put(node, dotNode);
				apply(node, dotNode, map, parameters);
			}
		}
		for (DirectedGraphEdge<? extends DirectedGraphNode, ? extends DirectedGraphNode> edge : graph.getEdges()) {
			addEdge(dot, map, parameters, nodeMap, graph, edge, (PetrinetNode) edge.getSource(), (PetrinetNode) edge.getTarget());
		}
		//		NavigableSVGPanel panel = new AnimatableSVGPanel(DotPanel.dot2svg(dot));
//		dot.setOption("splines", "polyline");
		dot.setDirection(GraphDirection.leftRight);
		return dot;
	}
	
	private void addEdge(Dot dot, ViewSpecificAttributeMap map, GraphVisualizerParameters parameters, Map<DirectedGraphNode, DotNode> nodeMap, DirectedGraph<? extends DirectedGraphNode, ? extends DirectedGraphEdge<?, ?>> graph, DirectedGraphEdge<?, ?> edge, PetrinetNode source, PetrinetNode target) {
		if (pnbpmnmap.hiddenNodes.contains(source) && !pnbpmnmap.hiddenNodes.contains(target)) {
			for (DirectedGraphEdge<? extends DirectedGraphNode, ? extends DirectedGraphNode> edge2 : graph.getInEdges(source)) {
				addEdge(dot, map, parameters, nodeMap, graph, edge2, (PetrinetNode) edge2.getSource(), target);
			}
		} else if (!pnbpmnmap.hiddenNodes.contains(source) && pnbpmnmap.hiddenNodes.contains(target)) {
			for (DirectedGraphEdge<? extends DirectedGraphNode, ? extends DirectedGraphNode> edge2 : graph.getOutEdges(target)) {
				addEdge(dot, map, parameters, nodeMap, graph, edge2, source, (PetrinetNode) edge2.getTarget());
			}
		} else if (!pnbpmnmap.hiddenNodes.contains(source) && !pnbpmnmap.hiddenNodes.contains(target)) {
			DotEdge dotEdge = dot.addEdge(nodeMap.get(source), nodeMap.get(target));
			apply(edge, dotEdge, map, parameters);	
		}
	}

	private void apply(DirectedGraphNode node, DotNode dotNode, ViewSpecificAttributeMap map,
			GraphVisualizerParameters parameters) {
		AttributeMap attMap = node.getAttributeMap();

		if (pnbpmnmap.andSplits.contains(node)) {
			dotNode.setOption("shape", "diamond");
			dotNode.setLabel("+");
			dotNode.setOption("fixedsize", "true");
			dotNode.setOption("height", "0.3");
			dotNode.setOption("width", "0.3");
			dotNode.setOption("fontsize", "13");
		} else if (pnbpmnmap.iorSplits.contains(node)) {
			dotNode.setOption("shape", "diamond");
			dotNode.setLabel("o");
			dotNode.setOption("fixedsize", "true");
			dotNode.setOption("height", "0.3");
			dotNode.setOption("width", "0.3");
			dotNode.setOption("fontsize", "13");
		} else if (pnbpmnmap.xorSplits.contains(node)) {
			dotNode.setOption("shape", "diamond");
			dotNode.setLabel("x");
			dotNode.setOption("fixedsize", "true");
			dotNode.setOption("height", "0.3");
			dotNode.setOption("width", "0.3");
			dotNode.setOption("fontsize", "13");
		} else {
			dotNode.setOption("shape", "box");
			Boolean showLabel = getBoolean(attMap, AttributeMap.SHOWLABEL, true, node, map);
			// HV: Setting a tooltip seems to have no effect.
			String label = getString(attMap, AttributeMap.LABEL, "", node, map);
			String placeLabel = getString(attMap, GVPLACELABEL, "", node, map);
			String tooltip = getString(attMap, AttributeMap.TOOLTIP, "", node, map);
			String internalLabel = getFormattedString(parameters.getInternalLabelFormat(), label, placeLabel, tooltip);
			String externalLabel = getFormattedString(parameters.getExternalLabelFormat(), label, placeLabel, tooltip);
			tooltip = getFormattedString(parameters.getToolTipFormat(), label, placeLabel, tooltip);
			if (showLabel) {
				dotNode.setLabel(internalLabel);
			} else {
				dotNode.setLabel("");
			}
			if (node instanceof Transition) {
				dotNode.setOption("xlabel", externalLabel);
			}
			dotNode.setOption("style", "rounded,filled");
			dotNode.setOption("fontsize", "12");
		}
		Float penWidth = getFloat(attMap, AttributeMap.LINEWIDTH, 1.0F, node, map);
		dotNode.setOption("penwidth", "" + penWidth);
		Color strokeColor = getColor(attMap, AttributeMap.STROKECOLOR, Color.BLACK, node, map);
		dotNode.setOption("color", ColourMap.toHexString(strokeColor));
		Color labelColor = getColor(attMap, AttributeMap.LABELCOLOR, Color.BLACK, node, map);
		dotNode.setOption("fontcolor", ColourMap.toHexString(labelColor));
		dotNode.setOption("fillcolor", ColourMap.toHexString(Color.WHITE));
	}

	/*
	 * Copy (as much as possible) the attributes from the JGraph edge to the dot
	 * edge.
	 */
	private void apply(DirectedGraphEdge<?, ?> edge, DotEdge dotEdge, ViewSpecificAttributeMap map,
			GraphVisualizerParameters parameters) {
		AttributeMap attMap = edge.getAttributeMap();
		Boolean showLabel = getBoolean(attMap, AttributeMap.SHOWLABEL, false, edge, map);
		String label = getString(attMap, AttributeMap.LABEL, "", edge, map);
		dotEdge.setLabel(showLabel ? label : "");
		dotEdge.setOption("dir", "both");
		ArrowType endArrowType = getArrowType(attMap, AttributeMap.EDGEEND, ArrowType.ARROWTYPE_CLASSIC, edge, map);
		Boolean endIsFilled = getBoolean(attMap, AttributeMap.EDGEENDFILLED, false, edge, map);
		switch (endArrowType) {
			case ARROWTYPE_SIMPLE :
			case ARROWTYPE_CLASSIC :
				dotEdge.setOption("arrowhead", "open");
				break;
			case ARROWTYPE_TECHNICAL :
				dotEdge.setOption("arrowhead", endIsFilled ? "normal" : "empty");
				break;
			case ARROWTYPE_CIRCLE :
				dotEdge.setOption("arrowhead", endIsFilled ? "dot" : "odot");
				break;
			case ARROWTYPE_LINE :
				dotEdge.setOption("arrowhead", "tee");
				break;
			case ARROWTYPE_DIAMOND :
				dotEdge.setOption("arrowhead", endIsFilled ? "diamond" : "odiamond");
				break;
			case ARROWTYPE_NONE :
				dotEdge.setOption("arrowhead", "none");
				break;
			default :
				dotEdge.setOption("arrowhead", endIsFilled ? "box" : "obox");
				break;
		}
		ArrowType startArrowType = getArrowType(attMap, AttributeMap.EDGESTART, ArrowType.ARROWTYPE_NONE, edge, map);
		Boolean startIsFilled = getBoolean(attMap, AttributeMap.EDGESTARTFILLED, false, edge, map);
		dotEdge.setOption("arrowtail", "none");
		switch (startArrowType) {
			case ARROWTYPE_SIMPLE :
			case ARROWTYPE_CLASSIC :
				dotEdge.setOption("arrowtail", "open");
				break;
			case ARROWTYPE_TECHNICAL :
				dotEdge.setOption("arrowtail", startIsFilled ? "normal" : "empty");
				break;
			case ARROWTYPE_CIRCLE :
				dotEdge.setOption("arrowtail", startIsFilled ? "dot" : "odot");
				break;
			case ARROWTYPE_LINE :
				dotEdge.setOption("arrowtail", "tee");
				break;
			case ARROWTYPE_DIAMOND :
				dotEdge.setOption("arrowtail", startIsFilled ? "diamond" : "odiamond");
				break;
			case ARROWTYPE_NONE :
				dotEdge.setOption("arrowtail", "none");
				break;
			default :
				dotEdge.setOption("arrowtail", startIsFilled ? "box" : "obox");
				break;
		}
		Float penWidth = getFloat(attMap, AttributeMap.LINEWIDTH, 1.0F, edge, map);
		dotEdge.setOption("penwidth", "" + penWidth);
		Color edgeColor = getColor(attMap, AttributeMap.EDGECOLOR, Color.BLACK, edge, map);
		dotEdge.setOption("color", ColourMap.toHexString(edgeColor));
		Color labelColor = getColor(attMap, AttributeMap.LABELCOLOR, Color.BLACK, edge, map);
		dotEdge.setOption("fontcolor", ColourMap.toHexString(labelColor));
	}

	/*
	 * The following methods get the attribute value from the JGraph object with
	 * a given default value. If the object has no such attribute, the default
	 * value will be returned.
	 */

	private Boolean getBoolean(AttributeMap map, String key, Boolean value, AttributeMapOwner owner,
			ViewSpecificAttributeMap m) {
		Object obj = m.get(owner, key, null);
		if (obj != null && obj instanceof Boolean) {
			return (Boolean) obj;
		}
		obj = map.get(key);
		if (obj != null && obj instanceof Boolean) {
			return (Boolean) obj;
		}
		return value;
	}

	private Float getFloat(AttributeMap map, String key, Float value, AttributeMapOwner owner,
			ViewSpecificAttributeMap m) {
		Object obj = m.get(owner, key, null);
		if (obj != null && obj instanceof Float) {
			return (Float) obj;
		}
		obj = map.get(key);
		if (obj != null && obj instanceof Float) {
			return (Float) obj;
		}
		return value;
	}

	private String getString(AttributeMap map, String key, String value, AttributeMapOwner owner,
			ViewSpecificAttributeMap m) {
		Object obj = m.get(owner, key, null);
		if (obj != null && obj instanceof String) {
			/*
			 * Some labels contain HTML mark-up. Remove as much as possible.
			 */
			String s1 = ((String) obj).replaceAll("<br>", "\\\\n");
			String s2 = s1.replaceAll("<[^>]*>", "");
			return s2;
		}
		obj = map.get(key);
		if (obj != null && obj instanceof String) {
			/*
			 * Some labels contain HTML mark-up. Remove as much as possible.
			 */
			String s1 = ((String) obj).replaceAll("<br>", "\\\\n");
			String s2 = s1.replaceAll("<[^>]*>", "");
			return s2;
		}
		return value;
	}

	private Color getColor(AttributeMap map, String key, Color value, AttributeMapOwner owner,
			ViewSpecificAttributeMap m) {
		Object obj = m.get(owner, key, null);
		if (obj != null && obj instanceof Color) {
			return (Color) obj;
		}
		obj = map.get(key);
		if (obj != null && obj instanceof Color) {
			return (Color) obj;
		}
		return value;
	}

	private ArrowType getArrowType(AttributeMap map, String key, ArrowType value, AttributeMapOwner owner,
			ViewSpecificAttributeMap m) {
		Object obj = m.get(owner, key, null);
		if (obj != null && obj instanceof ArrowType) {
			return (ArrowType) obj;
		}
		obj = map.get(key);
		if (obj != null && obj instanceof ArrowType) {
			return (ArrowType) obj;
		}
		return value;
	}

	private String getFormattedString(String format, String label, String placeLabel, String tooltip) {
		String shortPlaceLabel = (placeLabel.length() > 5 ? placeLabel.substring(0, 4) : placeLabel);
		return format.replace("%l", label).replace("%p", placeLabel).replace("%s", shortPlaceLabel).replace("%t",
				tooltip);
	}
}
