package org.processmining.prabhakardixit.processanalytics;

import java.awt.Font;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import prefuse.Constants;
import prefuse.Display;
import prefuse.Visualization;
import prefuse.action.ActionList;
import prefuse.action.RepaintAction;
import prefuse.action.assignment.ColorAction;
import prefuse.action.assignment.DataColorAction;
import prefuse.action.layout.StackedAreaChart;
import prefuse.controls.DragControl;
import prefuse.controls.PanControl;
import prefuse.controls.ZoomControl;
import prefuse.data.Schema;
import prefuse.data.Table;
import prefuse.data.expression.parser.ExpressionParser;
import prefuse.render.DefaultRendererFactory;
import prefuse.render.LabelRenderer;
import prefuse.render.PolygonRenderer;
import prefuse.util.ColorLib;
import prefuse.util.FontLib;
import prefuse.util.PrefuseLib;
import prefuse.visual.VisualItem;

public class StackedChart extends JFrame {
	static JScrollPane scrPane = new JScrollPane();

	public static void createStackedChart2(JFrame frame, Map<String, Map<Double, Double>> displayView) {
		JPanel contentPane = (JPanel) frame.getContentPane();
		contentPane.removeAll();
		contentPane.revalidate();
		contentPane.repaint();
		ActionList color = new ActionList();
		int[] palette = new int[] { ColorLib.rgba(255, 200, 200, 150), ColorLib.rgba(200, 255, 200, 150),
				ColorLib.rgba(255, 0, 0, 250) };
		ColorAction fillColor = new DataColorAction("table", "activity", Constants.NOMINAL, VisualItem.FILLCOLOR,
				palette);
		color.add(fillColor);
		int[] palette2 = new int[] { ColorLib.rgba(2, 21, 21, 15), ColorLib.rgba(2, 2, 2, 15),
				ColorLib.rgba(255, 0, 0, 250) };

		Set<Double> categories = getAllCategories(displayView);
		String[] fields = new String[categories.size()];
		int index = 0;
		for (Double category : categories) {
			fields[index] = String.valueOf(category);
			index++;
		}

		ActionList layout = new ActionList();
		layout.add(new RepaintAction());
		StackedAreaChart sac = new StackedAreaChart("table", VisualItem.POLYGON, fields);
		layout.add(sac);

		Visualization vis = new Visualization();
		Table table = new Table();
		vis.add("table", table);
		vis.addFocusGroup("focusgroup");
		//decorator starts
		((DefaultRendererFactory)vis.getRendererFactory()).add("INGROUP('labeldecorator')", new LabelRenderer("label")); 
		final Schema BAR_TEXT_DECORATOR_SCHEMA = PrefuseLib.getVisualItemSchema();
		BAR_TEXT_DECORATOR_SCHEMA.addColumn("label", String.class);
		BAR_TEXT_DECORATOR_SCHEMA.setDefault(VisualItem.INTERACTIVE, false);
		BAR_TEXT_DECORATOR_SCHEMA.setDefault(VisualItem.FONT, FontLib.getFont("Serif", Font.BOLD, 4));
		BAR_TEXT_DECORATOR_SCHEMA.setDefault(VisualItem.TEXTCOLOR, ColorLib.green(150));
		//decorator ends
		populateTable(displayView, table, categories);

		vis.putAction("layout", layout);
		vis.putAction("color", color);

		DefaultRendererFactory drf = new DefaultRendererFactory();
		drf.add("ingroup('table')", new PolygonRenderer());
		
		vis.setRendererFactory(drf);
		sac.getVisualization().addDecorators("labeldecorator", "table",
				ExpressionParser.predicate("[" + VisualItem.HIGHLIGHT + "] = TRUE"), BAR_TEXT_DECORATOR_SCHEMA); //decorator

		
		
		Display display = new Display(vis);
		display.setSize(720, 500);
		display.addControlListener(new DragControl());
		// pan with left-click drag on background
		display.addControlListener(new PanControl());
		// zoom with right-click drag
		display.addControlListener(new ZoomControl());
		display.addControlListener(new HoverControl());

		//		JFrame frame = new JFrame("Prefuse StackedAreaChart Example");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.add(display);
		frame.pack();
		frame.setVisible(true);
		
		sac.run();
		vis.run("layout");
		vis.run("color");
		//		return frame;
	}

	public static void createStackedChart3(JFrame frame, Map<String, Map<Double, Double>> displayView) {
		JPanel contentPane = (JPanel) frame.getContentPane();
		contentPane.removeAll();
		contentPane.revalidate();
		contentPane.repaint();
		ActionList color = new ActionList();
		int[] palette = new int[] { ColorLib.rgba(255, 200, 200, 150), ColorLib.rgba(200, 255, 200, 150) };
		ColorAction fillColor = new DataColorAction("table", "activity", Constants.NOMINAL, VisualItem.FILLCOLOR,
				palette);
		color.add(fillColor);

		Set<Double> categories = getAllCategories(displayView);
		String[] fields = new String[categories.size()];
		int index = 0;
		for (Double category : categories) {
			fields[index] = String.valueOf(category);
			index++;
		}
		Table table = new Table();
		populateTable(displayView, table, categories);

		StackedAreaChart sac = new StackedAreaChart("table", "name", fields);

		Visualization vis = new Visualization();
		vis.add("table", table);
		vis.addFocusGroup("focusgroup");
		//decorator starts
		final Schema BAR_TEXT_DECORATOR_SCHEMA = PrefuseLib.getVisualItemSchema();
		BAR_TEXT_DECORATOR_SCHEMA.setDefault(VisualItem.INTERACTIVE, false);
		BAR_TEXT_DECORATOR_SCHEMA.setDefault(VisualItem.FONT, FontLib.getFont("Serif", Font.BOLD, 4));
		BAR_TEXT_DECORATOR_SCHEMA.setDefault(VisualItem.TEXTCOLOR, ColorLib.getColor(250));
		//decorator ends

		vis.putAction("myChartLayout", sac);
		vis.putAction("color", color);
		vis.addDecorators("labeldecorator", "focusgroup", BAR_TEXT_DECORATOR_SCHEMA); //decorator
		Display display = new Display(vis);
		display.setSize(720, 500);
		display.addControlListener(new DragControl());
		// pan with left-click drag on background
		display.addControlListener(new PanControl());
		// zoom with right-click drag
		display.addControlListener(new ZoomControl());

		//		JFrame frame = new JFrame("Prefuse StackedAreaChart Example");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.add(display);
		frame.pack();
		frame.setVisible(true);
		sac.run();
		vis.run("myChartLayout");
		vis.run("color");
		//		return frame;
	}

	private static void populateTable(Map<String, Map<Double, Double>> displayView, Table table, Set<Double> categories) {

		table.addColumn("activity", String.class);
		for (Double category : categories) {
			table.addColumn(String.valueOf(category), float.class);
		}

		//		table.addColumn("1980s", int.class);
		//		table.addColumn("1990s", int.class);
		//		table.addColumn("2000s", int.class);
		table.addColumn(VisualItem.POLYGON, float[].class, null);
		table.addColumn(VisualItem.POLYGON + ":start", float[].class, null);
		table.addColumn(VisualItem.POLYGON + ":end", float[].class, null);

		for (String activity : displayView.keySet()) {
			int rowNumber = table.addRow();
			table.setString(rowNumber, "activity", activity);
			Set<Double> categoriesNotAdded = new HashSet<Double>(categories);
			for (Double category : displayView.get(activity).keySet()) {
				categoriesNotAdded.remove(category);
				table.setFloat(rowNumber, String.valueOf(category), displayView.get(activity).get(category)
						.floatValue());
			}
			for (Double category : categoriesNotAdded) {
				table.setFloat(rowNumber, String.valueOf(category), 0);
			}
		}
		//		
		//		int rowNumber = table.addRow();
		//		table.setString(rowNumber, "activity", "Bob");
		//		table.setInt(rowNumber, "1980s", 12000);
		//		table.setInt(rowNumber, "1990s", 1500);
		//		table.setInt(rowNumber, "2000s", 1300);
		//
		//		rowNumber = table.addRow();
		//		table.setString(rowNumber, "activity", "Mary");
		//		table.setInt(rowNumber, "1980s", 800);
		//		table.setInt(rowNumber, "1990s", 1500);
		//		table.setInt(rowNumber, "2000s", 3200);

	}

	private static SortedSet<Double> getAllCategories(Map<String, Map<Double, Double>> displayView) {
		//		Set<Double> categories = new HashSet<Double>();
		SortedSet<Double> categories = new TreeSet<>();
		for (String key : displayView.keySet()) {
			for (Double val : displayView.get(key).keySet()) {
				categories.add(val);
			}
		}
		return categories;
	}

}